# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
PyTest suite for testing all module specification:
"""
import json
import os
import logging
import traceback

from shrike.pipeline import AMLPipelineHelper


log = logging.getLogger(__name__)


def get_config_class(pipeline_class):
    """Test if the get_arg_parser() method is in there and behaves correctly"""

    try:
        config_class = pipeline_class.get_config_class()
    except:
        assert (
            False
        ), "getting config class for pipeline class {} resulted in an exception: {}".format(
            pipeline_class.__name__, traceback.format_exc()
        )


def pipeline_required_modules(pipeline_class):
    """Test if the required_modules() returns the right list of modules with all required keys"""
    modules_manifest = pipeline_class.required_modules()

    assert isinstance(
        modules_manifest, dict
    ), "required_modules() must return a dictionary."

    error_log = []
    for module_key, module_description in modules_manifest.items():
        if not isinstance(module_description, dict):
            error_log.append(
                f"values in dictionary returned by required_modules() must be dictionaries (under key={module_key}, found value of type={module_description.__name__})"
            )
            continue
        if "yaml_spec" not in module_description:
            error_log.append(
                f"In pipeline class module {pipeline_class.__name__},"
                + f" module under key={module_key} (in required_modules() function)"
                + " does not provide any yaml_spec key."
                + " You need to give such a yaml_spec path before creating your pull request"
                + " so that we're able to consume this module when running pre-merge tests (detonation chamber)"
            )
        if "remote_module_name" not in module_description:
            error_log.append(
                f"In pipeline class module {pipeline_class.__name__},"
                + f" module under key={module_key} (in required_modules() function)"
                + " does not provide any remote_module_name."
                + " You need to give such a name before creating your pull request"
                + " so that we're able to consume this module when running in production."
            )

    assert not (error_log), (
        f"In pipeline class module {pipeline_class.__name__}, validation of the dictionary returned by required_modules() method shows errors:\n"
        + "\n".join(error_log)
    )


def pipeline_required_subgraphs(pipeline_class):
    """Tests if the required_subgraphs() returns the right list of modules with all requires keys"""
    subgraphs_manifest = pipeline_class.required_subgraphs()

    assert isinstance(
        subgraphs_manifest, dict
    ), "required_subgraphs() must return a dictionary."

    error_log = []
    for subgraph_key, subgraph_class in subgraphs_manifest.items():
        if not issubclass(subgraph_class, AMLPipelineHelper):
            error_log.append(
                f"In pipeline class module {pipeline_class.__name__}, values in dictionary returned by required_subgraphs() must be subclass of AMLPipelineHelper (under key={subgraph_key}, found object {subgraph_class.__name__})"
            )
            continue

    assert not (error_log), (
        f"In pipeline class module {pipeline_class.__name__}, validation of the dictionary returned by required_subgraphs() shows errors: "
        + "\n".join(error_log)
    )


def deeptest_graph_comparison(pipeline_export_file, pipeline_definition_file):
    """Compare a pipeline object to a serialized definition [EXPERIMENTAL]

    Args:
        pipeline_export_file (str): path to pipeline exported file
        pipeline_definition_file (str): path to reference file

    Returns:
        None
    """
    # checks the exported file in temp dir
    assert os.path.isfile(
        pipeline_export_file
    ), f"deeptest_graph_comparison() expects a file as first argument but {pipeline_export_file} does not exist."
    assert os.path.isfile(
        pipeline_definition_file
    ), f"deeptest_graph_comparison() expects a file as second argument but {pipeline_definition_file} does not exist."

    # read the exported graph
    with open(pipeline_export_file, "r") as export_file:
        pipeline = json.loads(export_file.read())
    assert (
        pipeline is not None
    ), f"deeptest_graph_comparison() expects first argument to be a parsable json, instead it found None"

    with open(pipeline_definition_file, "r") as definition_file:
        definition = json.loads(definition_file.read())
    assert (
        definition is not None
    ), f"deeptest_graph_comparison() expects first argument to be a parsable json, instead it found None"

    deeptest_graph(pipeline, definition)


def deeptest_graph(pipeline, definition, path="ROOT"):
    """Recursively compare a pipeline object to a serialized definition [EXPERIMENTAL]

    Args:
        pipeline (json): source for the comparison
        definition (json): target/reference for the comparison
        path (str): current path of the comparison (in the json tree)

    Returns:
        None
    """
    if definition is None:
        # no definition provided, let's stop inspection at this path
        log.info(f"deeptest_graph @ {path}: nop, definition is None")
        return

    # is inspecting a dictionary structure, iterate on keys
    if isinstance(pipeline, dict) and isinstance(definition, dict):
        log.info(f"deeptest_graph @ {path}: checking dictionary")
        for key in definition:
            assert (
                key in pipeline
            ), f"pipeline graph does not have key {key} at level @ {path}"

            # ignoring all ids
            if key in {"id", "node_id", "module_id", "dataset_id"}:
                log.info(f"deeptest_graph @ {path}: ignore id key {key}")
                return

            if (
                key in {"run_settings", "compute_run_settings"}
                and definition[key] is not None
            ):
                # this is a specific kind of key containing a list we're transforming into a dict
                log.info(f"deeptest_graph @ {path}: refactoring key {key} as dict")
                pipeline_run_settings = dict(
                    [(entry["name"], entry) for entry in pipeline[key]]
                )
                definition_run_settings = dict(
                    [(entry["name"], entry) for entry in definition[key]]
                )
                deeptest_graph(
                    pipeline_run_settings,
                    definition_run_settings,
                    path + ".(runsettings)" + key,
                )
            else:
                deeptest_graph(pipeline[key], definition[key], path + "." + key)
        return

    # is inspecting a list structure, each element MUST passed
    # NOTE: this should be improved in case list can be shuffled ?
    if isinstance(pipeline, list) and isinstance(definition, list):
        log.info(f"deeptest_graph @ {path}: checking list")
        for key, entry in enumerate(definition):
            deeptest_graph(pipeline[key], entry, path + "[" + str(key) + "]")
        return

    # if anything else (int, str, unknown), just test plain equality
    log.info(f"deeptest_graph @ {path}: checking equality {pipeline} == {definition}")
    assert pipeline == definition, f"values mismatch @ {path}"



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Helper code to run Dask distributed scripts [EXPERIMENTAL]
"""
import asyncio
import os
import socket
import sys
from typing import Optional, cast

import psutil
from shrike._core import experimental
from shrike.compliant_logging import DataCategory
from shrike.distributed import EXPERIMENTAL_WARNING_MSG
from shrike.distributed.cluster_auto_setup import ClusterAutoSetupHandler

from dask.distributed import Client, Nanny, Scheduler, Worker, performance_report


# pragma: no cover
class DaskClusterSetupHandler(ClusterAutoSetupHandler):
    def __init__(
        self,
        process_count_per_node: Optional[int] = None,
        thread_count_per_process: Optional[int] = None,
        dask_report_link: Optional[str] = None,
        nanny: bool = False,
        public_log: bool = False,
    ):
        """Constructor for Dask Cluster
        Args:
            process_count_per_node (int): Number of processes available per node.
                Defaults to `None`.
                If `None` the machine cpu count defines the number of processes.
            thread_count_per_process (int): Number of threads available per process.
                Defaults to `None`. If `None` the number of threads is defined by the
                division of cpu count by number of processes.
            dask_report_link (str): Path where dask dashboard report will be written to.
                Defaults to None. When None, dask dashboard is not reported/written to
                a file.
            Nanny (bool): Whether to use Nanny or not. Defaults to False.
                The nanny spins up Worker processes, watches then, and kills or restarts
                them as necessary. It is necessary if you want to use
                the Client.restart method, or to restart the worker automatically if
                it gets to the terminate fraction of its memory limit
            public_log (bool): Whether to log information to the compliant public log.
                If used call `shrike.compliant_logging.enable_compliant_logging()` to
                initialize compliant logging beforehand. Defaults to False.

        Example:
            - Starts a Cluster using as many processes as cpu count
                (Ideal to avoid workloads suffering from GIL lock):

            >>> from shrike.distributed import DaskClusterSetupHandler
            >>> with DaskClusterSetupHandler() as azure_cluster:
            ...     client = azure_cluster.client
            ...     logger.info(f"Dask {client}")

            - Starts a Cluster using 1 process per machine,
                and give to it as many threads as cpu count:

            >>> from shrike.distributed import DaskClusterSetupHandler
            >>> with DaskClusterSetupHandler(process_count_per_node=1) as azure_cluster:
            ...     client = azure_cluster.client
            ...     logger.info(f"Dask {client}")
        """
        super().__init__()
        self.public_log = public_log
        self.head_port = 45462
        if nanny is True:
            self.constructor = Nanny
        else:
            self.constructor = Worker

        self.node_count = int(cast(str, os.getenv("OMPI_COMM_WORLD_SIZE")))
        self.rank = int(cast(str, os.getenv("OMPI_COMM_WORLD_RANK")))
        self.ip = socket.gethostbyname(socket.gethostname())

        self.head_address = None
        if process_count_per_node is None:
            self.process_count_per_node = psutil.cpu_count()
        else:
            self.process_count_per_node = process_count_per_node

        # How many threads is available per worker
        if thread_count_per_process is None:
            self.thread_count_per_process = int(
                psutil.cpu_count() / self.process_count_per_node
            )
        else:
            self.thread_count_per_process = int(thread_count_per_process)
        self.worker_ram = int(
            0.90 * psutil.virtual_memory().available / self.process_count_per_node
        )
        self.worker_count = (self.node_count - 1) * self.process_count_per_node

        self.dask_performance_report_file: Optional[str] = None
        # output dashboard
        if dask_report_link is not None:
            self.dask_performance_report_file = os.path.join(
                dask_report_link, "dask-report.html"
            )

    def _log_info_publicly(self, message):
        """Logs a message to the public log.
        When compliant logging is enabled, this will log to the public log.
        Otherwise, this will log to info
        """
        if self.public_log:
            self.logger.info(message, category=DataCategory.PUBLIC)
        else:
            self.logger.info(message)

    #################
    # SETUP METHODS #
    #################

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def setup_head_node(self):
        """Setup to run only on head node"""
        self._log_info_publicly(EXPERIMENTAL_WARNING_MSG)
        self._log_info_publicly(
            f"{self.__class__.__name__}.setup_head_node() called to set up HEAD node."
        )
        # create setup config
        self.head_address = socket.gethostbyname(socket.gethostname())

        # record what's needed to setup cluster nodes
        self.setup_config_add_key("head_address", self.head_address)

        self.pid = os.fork()  # type: ignore
        if self.pid > 0:
            self.rank = "0.1"

            async def run_scheduler():
                async with Scheduler(
                    port=self.head_port, scheduler_file="scheduler.json"
                ) as scheduler:
                    await scheduler.finished()

            asyncio.get_event_loop().run_until_complete(run_scheduler())
            self._log_info_publicly(f"[R{self.rank}]SCHEDULER IS OFFLINE")
            self.multinode_driver.finalize()
            sys.exit(0)

    @experimental(message=EXPERIMENTAL_WARNING_MSG)
    def setup_cluster_node(self):
        """Setup to run only on non-head cluster nodes"""
        self._log_info_publicly(EXPERIMENTAL_WARNING_MSG)

        self._log_info_publicly(
            f"{self.__class__.__name__}.setup_cluster_node() called"
        )
        self.head_address = self.setup_config_get_key("head_address")

    def __enter__(self):
        self.initialize_run()

        for line in self.__repr__().split("\n"):
            self._log_info_publicly(line)

        if self.rank != 0:
            self._log_info_publicly("Running worker...")

            async def run_worker():
                async with self.constructor(
                    scheduler_ip=self.head_address,
                    scheduler_port=self.head_port,
                    nthreads=self.thread_count_per_process,
                    memory_limit=self.worker_ram,
                    local_directory=os.getenv("AZ_BATCHAI_JOB_TEMP"),
                    name=f"{self.rank}",
                ) as worker:
                    self._log_info_publicly(f"[R{self.rank}] - {worker}")
                    await worker.finished()

            if self.process_count_per_node > 1:
                for i in range(self.process_count_per_node - 1):
                    pid = os.fork()
                    if pid > 0:
                        self.rank = f"{self.rank}_{i}"
                        asyncio.get_event_loop().run_until_complete(run_worker())
                        self._log_info_publicly(f"[R{self.rank}]Worker IS OFFLINE")
                        self.multinode_driver.finalize()
                        sys.exit(0)

            # Shut down when down
            asyncio.get_event_loop().run_until_complete(run_worker())
            self._log_info_publicly(f"[R{self.rank}]Worker IS OFFLINE")
            self.multinode_driver.finalize()
            sys.exit(0)
        else:
            self._log_info_publicly("Starting Dask Client...")
            self.client = Client(f"tcp://{self.head_address}:{self.head_port}")
            self._log_info_publicly(
                f"Waiting for {self.worker_count} workers up to 10s"
            )
            self.client.wait_for_workers(self.worker_count, 10)

            if self.dask_performance_report_file is not None:
                self.dask_performance_report = performance_report(
                    filename=self.dask_performance_report_file
                )
                self.dask_performance_report.__enter__()

        return self

    def __repr__(self) -> str:
        string = (
            "Cluster info\n"
            f"\t├─>\t\t rank = {self.rank} \n"
            f"\t├─>\t\t ip = {self.ip} \n"
            f"\t├─>\t\t scheduler_ip = {self.head_address} \n"
            f"\t├─>\t\t node_count = {self.node_count} \n"
            f"\t├─>\t\t process_count_per_node = {self.process_count_per_node} \n"
            f"\t├─>\t\t thread_count_per_process = {self.thread_count_per_process} \n"
            f"\t├─>\t\t worker_ram = {(self.worker_ram/(2**30)):.2f} GB \n"
            f"\t├─>\t\t worker_count = {self.worker_count} \n"
        )
        return string

    def __exit__(self, type, value, traceback):

        self._log_info_publicly(f"Exiting context from rank {self.rank}")
        if self.rank == 0:

            self._log_info_publicly(
                f"Dask dashboard saved to {self.dask_performance_report_file}"
            )
            self._log_info_publicly("Shutting dask down...")
            self.client.shutdown()  # This will also close the workers

            if os.path.exists("scheduler.json"):
                os.remove("scheduler.json")
            # self.multinode_driver.finalize()
            if self.dask_performance_report_file is not None:
                self.dask_performance_report.__exit__(type, value, traceback)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
from abc import abstractmethod, ABC
from dataclasses import dataclass, field
import os
import logging
from typing import Optional, List, Tuple

log = logging.getLogger(__name__)


@dataclass
class module_reference:
    key: Optional[str] = (
        None  # use as internal key to reference module (if None, use name)
    )
    name: Optional[str] = None  # None if module exists only locally?
    source: Optional[str] = None  # This config is deprecated. Please do not use.
    yaml: Optional[str] = None
    version: Optional[str] = None
    from_workspace: Optional[bool] = (
        False  # If set to True and it's a remote component, shrike will only attempt to load it from workspace (not from registry or feed)
    )


@dataclass
class module_manifest:
    manifest: List[module_reference] = field(default_factory=list)
    feeds: List[str] = field(default_factory=list)
    registries: List[str] = field(default_factory=list)


@dataclass
class module_loader_config:  # pylint: disable=invalid-name
    """Config for the AMLModuleLoader class"""

    use_local: Optional[str] = None
    use_local_except_for: Optional[bool] = None
    force_default_module_version: Optional[str] = None
    force_all_module_version: Optional[str] = None
    local_steps_folder: str = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "..",
            "tests",
            "tests_pipeline",
            "sample",
            "steps",
        )
    )
    use_remote_when_component_not_in_manifest: Optional[bool] = False
    loading_order: Optional[str] = "registry, feed, workspace"


class AMLModuleLoaderBase(ABC):
    """Helper class to load modules from within an AMLPipelineHelper."""

    def __init__(self, config):
        """Creates module instances for AMLPipelineHelper.

        Args:
            config (DictConfig): configuration options
        """
        self.use_local_except_for = (
            config.module_loader.use_local_except_for
            if "use_local_except_for" in config.module_loader
            else None
        )
        if "use_local" not in config.module_loader:
            self.use_local = []
        elif config.module_loader.use_local is None:
            self.use_local = []
        elif config.module_loader.use_local == "*":
            self.use_local = "*"
        elif isinstance(config.module_loader.use_local, str):
            self.use_local = [
                x.strip() for x in config.module_loader.use_local.split(",")
            ]
            if not self._check_use_local_syntax_valid(self.use_local):
                raise ValueError(
                    f'Invalid value for `use_local`. Please follow one of the four patterns: \n1) use_local="", all modules are remote\n2) use_local="*", all modules are local\n3) use_local="MODULE_KEY_1, MODULE_KEY_2", only MODULE_KEY_1, MODULE_KEY_2 are local, everything else is remote\n4) use_local="!MODULE_KEY_1, !MODULE_KEY_2", all except for MODULE_KEY_1, MODULE_KEY_2 are local'
                )
            self.use_local_except_for = self.use_local[0].startswith("!")

        self.force_default_module_version = (
            config.module_loader.force_default_module_version
            if "force_default_module_version" in config.module_loader
            else None
        )
        self.force_all_module_version = (
            config.module_loader.force_all_module_version
            if "force_all_module_version" in config.module_loader
            else None
        )
        self.local_steps_folder = config.module_loader.local_steps_folder
        self.use_remote_when_component_not_in_manifest = (
            config.module_loader.use_remote_when_component_not_in_manifest
            if "use_remote_when_component_not_in_manifest" in config.module_loader
            else None
        )
        self.module_cache = {}

        self.feeds: List[str] = []
        self.registries: List[str] = []

        self.loading_order = (
            [x.strip().lower() for x in config.module_loader.loading_order.split(",")]
            if "loading_order" in config.module_loader
            else ["registry", "feed", "workspace"]
        )

        # internal manifest built from yaml config
        self.modules_manifest = {}
        self.load_config_manifest(config)

        initialization_info_string = (
            "AMLModuleLoader initialized ("
            f"use_local={self.use_local}"
            f", force_default_module_version={self.force_default_module_version}"
            f", force_all_module_version={self.force_all_module_version}"
            f", local_steps_folder={self.local_steps_folder}"
            f", use_remote_when_component_not_in_manifest={self.use_remote_when_component_not_in_manifest}"
            f", manifest={list(self.modules_manifest.keys())}"
            f", loading_order={self.loading_order}"
            ")"
        )
        log.info(initialization_info_string)

    @abstractmethod
    def solve_module_version_and_load(
        self, module_name, module_version, module_cache_key, registry=None
    ):
        """Loads module class if exists

        Args:
            module_name (str): name of the module to load
            module_version (str): version of the module to load
            module_cache_key (str): cache key of the module after loading
            registy (str): registry name if loading from a registry
        """
        pass

    def load_config_manifest(self, config):
        """Fills the internal module manifest based on config object"""
        for entry in config.modules.manifest:
            if entry.key:
                module_key = entry.key
            elif entry.name:
                module_key = entry.name
            else:
                raise Exception(
                    "In module manifest, you have to provide at least key or name."
                )

            self.modules_manifest[module_key] = entry

        for feed in config.modules.get("feeds", []):
            self.feeds.append(feed)

        for registry in config.modules.get("registries", []):
            self.registries.append(registry)

    def is_local(self, module_name):
        """Tests is module is in local list"""
        if self.use_local == "*":
            return True
        if self.use_local_except_for:
            return "!" + module_name not in self.use_local
        else:
            return module_name in self.use_local

    def module_in_cache(self, module_cache_key):
        """Tests if module in internal cache (dict)"""
        return module_cache_key in self.module_cache

    def get_from_cache(self, module_cache_key):
        """Gets module class from internal cache (dict)"""
        log.debug(f"Using cached module {module_cache_key}")
        return self.module_cache.get(module_cache_key, None)

    def put_in_cache(self, module_cache_key, module_class):
        """Puts module class in internal cache (dict)"""
        self.module_cache[module_cache_key] = module_class

    def verify_manifest(self, modules_manifest):
        """Tests a module manifest schema"""
        errors = []

        for k, module_entry in modules_manifest.items():
            # TODO: merge error checking code with processing code so we do all this in one pass
            if self.is_local(k):
                if "yaml_spec" not in module_entry:
                    errors.append(
                        f"{k}: You need to specify a yaml_spec for your module to use_local=['{k}']"
                    )
                elif not os.path.isfile(
                    module_entry["yaml_spec"]
                ) and not os.path.isfile(
                    os.path.join(self.local_steps_folder, module_entry["yaml_spec"])
                ):
                    errors.append(
                        "{}: Could not find yaml spec {} for use_local=['{}']".format(
                            k, module_entry["yaml_spec"], k
                        )
                    )
            else:
                if "remote_module_name" not in module_entry:
                    errors.append(
                        f"{k}: You need to specify a name for your module to use_local=False"
                    )
                if "namespace" not in module_entry:
                    errors.append(
                        f"{k}: You need to specify a namespace for your module to use_local=False"
                    )
                if ("version" not in module_entry) and (
                    self.force_default_module_version or self.force_all_module_version
                ):
                    errors.append(
                        f"{k}: You need to specify a version for your module to use_local=False, or use either force_default_module_version or force_all_module_version in config"
                    )

        return errors

    @abstractmethod
    def load_local_module_helper(self, module_spec_path):
        """Calls SDK to load a local component.

        Args:
            module_spec_path (str): path to local module yaml spec

        Returns:
            object: module class loaded"""
        pass

    def load_local_module(self, module_spec_path):
        """Creates one module instance.

        Args:
            module_spec_path (str): path to local module yaml spec

        Returns:
            object: module class loaded
        """
        module_cache_key = module_spec_path
        if self.module_in_cache(module_cache_key):
            return self.get_from_cache(module_cache_key)

        log.info("Building module from local code at {}".format(module_spec_path))
        if not os.path.isfile(module_spec_path):
            module_spec_path = os.path.join(self.local_steps_folder, module_spec_path)
        loaded_module_class = self.load_local_module_helper(module_spec_path)
        self.put_in_cache(module_cache_key, loaded_module_class)

        return loaded_module_class

    def load_prod_module_helper(
        self,
        module_name,
        module_version,
        module_cache_key,
        module_namespace=None,
        loading_order=None,
    ):
        """Loads module from available sources

        Args:
            module_name (str): name of the module to load
            module_version (str): version of the module to load
            module_cache_key (str): cache key of the module after loading
            module_namespace (str): namespace of the module if applicable
            loading_order (list[str]): custom loading order, if not specified loading order from config is used
        """

        if loading_order is None:
            loading_order = self.loading_order

        for source in loading_order:
            if source == "registry":
                log.info(f"Attempting to load {module_name} from registries...")
                for registry in self.registries:
                    try:
                        log.info(
                            f"Loading component {module_name} from registry {registry}"
                        )

                        loaded_module_class = self.solve_module_version_and_load(
                            module_name, module_version, module_cache_key, registry
                        )
                        return loaded_module_class

                    except BaseException as e:
                        log.debug(f"Exception: {e}")

            elif source == "feed":
                log.info(f"Attempting to load {module_name} from feeds...")
                for feed in self.feeds:
                    try:
                        log.info(f"Loading component {module_name} from feed {feed}")

                        module_name = f"azureml.feed://{feed}/{module_name}"

                        loaded_module_class = self.solve_module_version_and_load(
                            module_name, module_version, module_cache_key
                        )
                        return loaded_module_class

                    except BaseException as e:
                        log.debug(f"Exception: {e}")

            elif source == "workspace":
                log.info(f"Attempting to load {module_name} from workspace...")
                try:
                    # try without namespace first
                    return self.solve_module_version_and_load(
                        module_name, module_version, module_cache_key
                    )
                except BaseException as e:
                    log.debug(f"Exception: {e}")

                if module_namespace is not None:
                    log.info(
                        f"    Trying to load module {module_name} with namespace {module_namespace}."
                    )
                    module_name = module_namespace + "://" + module_name

                    try:
                        return self.solve_module_version_and_load(
                            module_name, module_version, module_cache_key
                        )
                    except BaseException as e:
                        log.debug(f"Exception: {e}")

            else:
                raise ValueError(
                    f"Loading order contains incorrect key {source}. Please use a comma-separated str to order 'registry', 'feed', and 'workspace'."
                )

        raise ValueError(
            f"Module {module_name} with version {module_version} not found in neither of the following sources: {loading_order}."
        )

    def load_prod_module(
        self, module_name, module_version, from_workspace=False, module_namespace=None
    ):
        """Creates one module instance.

        Args:
            module_name (str) : module name
            module_version (str) : module version
            from_workspace (bool) : attempt to load from workspace only

        Returns:
            object: module class loaded
        """
        if self.force_all_module_version:
            module_version = self.force_all_module_version
        else:
            module_version = module_version or self.force_default_module_version

        module_cache_key = f"{module_name}:{module_version}"
        if self.module_in_cache(module_cache_key):
            return self.get_from_cache(module_cache_key)

        log.info(
            f"Loading remote module {module_cache_key} (name={module_name}, version={module_version}, namespace={module_namespace})"
        )

        if from_workspace:
            return self.load_prod_module_helper(
                module_name,
                module_version,
                module_cache_key,
                module_namespace,
                loading_order=["workspace"],
            )
        return self.load_prod_module_helper(
            module_name, module_version, module_cache_key, module_namespace
        )

    def get_module_manifest_entry(
        self, module_key, modules_manifest=None
    ) -> Tuple[dict, Optional[str], bool]:
        """Gets a particular entry in the module manifest.

        Args:
            module_key (str): module key from the manifest
            modules_manifest (dict): manifest from required_modules() [DEPRECATED]

        Returns:
            module_entry (dict): module manifest entry (if no entry for this module key in the manifest, only the module key is returned)
            module_namespace (str | None): module namespace for legacy modules
            is_in_manifest (bool): true if the module key can be found in the manifest
        """
        if module_key in self.modules_manifest:
            module_entry = self.modules_manifest[module_key]
            module_namespace = None
            is_in_manifest = True
        elif modules_manifest and module_key in modules_manifest:
            log.warning(
                f"We highly recommend substituting the `required_modules` method by the modules.manifest configuration."
            )
            module_entry = modules_manifest[module_key]
            # map to new format
            module_entry["yaml"] = module_entry["yaml_spec"]
            module_entry["name"] = module_entry["remote_module_name"]
            module_namespace = module_entry.get("namespace", None)
            is_in_manifest = True
        else:
            module_entry = {}
            module_entry["name"] = module_key
            module_namespace = None
            is_in_manifest = False
            if not (self.use_remote_when_component_not_in_manifest):
                log.warning(
                    f"Module key '{module_key}' could not be found in modules.manifest configuration or in required_modules() method. If you want to try and load it from the workspace, set 'module_loader.use_remote_when_component_not_in_manifest' to True in the config."
                )

        return module_entry, module_namespace, is_in_manifest

    def load_module(self, module_key, modules_manifest=None):
        """Loads a particular module from the manifest.

        Args:
            module_key (str): module key from the manifest
            modules_manifest (dict): manifest from required_modules() [DEPRECATED]

        Returns:
            object: module class loaded
        """
        module_entry, module_namespace, is_in_manifest = self.get_module_manifest_entry(
            module_key, modules_manifest
        )
        if is_in_manifest:
            if self.is_local(module_key):
                loaded_module = self.load_local_module(module_entry["yaml"])
            else:
                loaded_module = self.load_prod_module(
                    module_entry["name"],
                    module_entry["version"],
                    from_workspace=(
                        module_entry["from_workspace"]
                        if "from_workspace" in module_entry
                        else False
                    ),
                    module_namespace=module_namespace,
                )
        else:
            log.warning(
                f"The component '{module_key}' cannot be found in the manifest. Attempting to load the remote copy."
            )
            loaded_module = self.load_prod_module(module_entry["name"], None)
        return loaded_module

    def load_modules_manifest(self, modules_manifest):
        """Creates module instances from modules_manifest.

        Args:
            modules_manifest (dict): manifest of modules to load

        Returns:
            dict: modules loaded, keys are taken from module_manifest.

        Raises:
            Exception: if loading module has an error or manifest is wrong.
        """
        log.info(f"Loading module manifest (use_local={self.use_local})")
        test_results = self.verify_manifest(modules_manifest)
        if test_results:
            raise Exception(
                "Loading modules from manifest raised errors:\n\nMANIFEST: {}\n\nERRORS: {}".format(
                    modules_manifest, "\n".join(test_results)
                )
            )

        loaded_modules = {}
        for module_key in modules_manifest:
            log.info(f"Loading module {module_key} from manifest")
            loaded_modules[module_key] = self.load_module(module_key, modules_manifest)

        return loaded_modules

    def _check_use_local_syntax_valid(self, use_local_list) -> bool:
        use_local_except_for = True if use_local_list[0].startswith("!") else False
        if use_local_except_for:
            for module_key in use_local_list:
                if not module_key.startswith("!"):
                    return False
        else:
            for module_key in use_local_list:
                if module_key.startswith("!"):
                    return False
        return True



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Constant values used by this library.
"""

from azure.ai.ml._logging.compliant_logger import DataCategory  # noqa

# Keep the file path for backwards compatibility,
# but import the class directly from AML SDK V2 and expose it here
__all__ = ["DataCategory"]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
import argparse
from abc import ABC, abstractmethod
import os
import sys
import json
import logging
import re
import webbrowser
import shutil
import yaml
import jsonpath_ng
from typing import Callable
from dataclasses import dataclass
from functools import lru_cache

try:
    import hydra
    from hydra.core.config_store import ConfigStore
    from hydra.core.hydra_config import HydraConfig
    from omegaconf import DictConfig, OmegaConf, open_dict

except ImportError as error:
    raise ImportError(
        f"{error.msg}. Please install using `pip install shrike[pipeline]`."
    )

from shrike import __version__
from shrike.pipeline.canary_helper import get_repo_info
from shrike.pipeline.pipeline_config import (
    default_config_dict,
)
from shrike.pipeline.telemetry_utils import TelemetryLogger
from shrike.pipeline.exceptions import ShrikeUserErrorException
from shrike._core import POLYMER_FEED, O365_FEED

log = logging.getLogger(__name__)


@dataclass
class Data:
    datastore_name: str
    path: str


class AMLPipelineHelperBase(ABC):
    """Helper class for building pipelines"""

    BUILT_PIPELINE = None  # the hydra run decorator doesn't allow for return, we're using this variable instead (hack)

    def __init__(self, config, module_loader):
        """Constructs the pipeline helper.

        Args:
            config (DictConfig): config for this object
            module_loader (AMLModuleLoader): which module loader to (re)use
        """
        self.repository_info = {}
        self.config = config
        self.module_loader = module_loader

        self.unified_workspace = (
            "uw_config" in self.config and len(self.config.uw_config.computes) > 0
        )

        if self.unified_workspace:
            self.config.compute = self.config.uw_config
            self.config.run.skip_update_dc = True  # TODO: without this, all outputs will be set to default_datastore (per silo, dts_via_cosmos)

    ######################
    ### CUSTOM METHODS ###
    ######################

    @classmethod
    def get_config_class(cls):
        """Returns a dataclass containing config for this pipeline"""
        pass

    @classmethod
    def required_subgraphs(cls):
        """Dependencies on other subgraphs
        Returns:
            dict[str, AMLPipelineHelper]: dictionary of subgraphs used for building this one.
                keys are whatever string you want for building your graph
                values should be classes inherinting from AMLPipelineHelper.
        """
        return {}

    @classmethod
    def required_modules(cls):
        """Dependencies on modules/components

        Returns:
            dict[str, dict]: manifest
        """
        return {}

    def build(self, config):
        """Builds a pipeline function for this pipeline.

        Args:
            config (DictConfig): configuration object (see get_config_class())

        Returns:
            pipeline_function: the function to create your pipeline
        """
        raise NotImplementedError("You need to implement your build() method.")

    def extra_tags(self):
        """Additional tags to submit with the pipeline

        Returns:
            dict[str, str]: tags
        """
        return {}

    def pipeline_instance(self, pipeline_function, config):
        """Creates an instance of the pipeline using arguments.

        Args:
            pipeline_function (function): the pipeline function obtained from self.build()
            config (DictConfig): configuration object (see get_config_class())

        Returns:
            pipeline: the instance constructed using build() function
        """
        raise NotImplementedError(
            "You need to implement your pipeline_instance() method."
        )

    @abstractmethod
    def _get_job_status(self, job):
        """Gets the job status"""
        pass

    @abstractmethod
    def _wait_for_job_completion(self, job):
        """Waits for the completion of this job"""
        pass

    @abstractmethod
    def _publish_to_endpoint(self, pipeline):
        """Publishes pipeline as endpoint"""
        pass

    def canary(self, args, experiment, pipeline_run):
        """Tests the output of the pipeline"""
        pass

    ##################################
    ### USER FACING HELPER METHODS ###
    ##################################

    @abstractmethod
    @lru_cache()
    def ml_client(self):
        """Gets the current workspace"""
        pass

    @abstractmethod
    @lru_cache()
    def component_load(self, component_key) -> Callable[..., "Component"]:
        """Loads one component from the manifest"""
        pass

    @abstractmethod
    @lru_cache()
    def module_load(self, module_key):
        """Loads one module from the manifest"""
        pass

    @lru_cache()
    def subgraph_load(self, subgraph_key, custom_config=OmegaConf.create()) -> Callable:
        """Loads one subgraph from the manifest
        Args:
            subgraph_key (str): subgraph identifier that is used in the required_subgraphs() method
            custom_config (DictConfig): custom configuration object, this custom object will be
            added to the pipeline config

        """
        subgraph_class = self.required_subgraphs()[subgraph_key]

        subgraph_config = self.config.copy()
        if custom_config:
            with open_dict(subgraph_config):
                for key in custom_config:
                    subgraph_config[key] = custom_config[key]

        log.info(f"Building subgraph [{subgraph_key} as {subgraph_class.__name__}]...")
        # NOTE: below creates subgraph with updated pipeline_config
        subgraph_instance = subgraph_class(
            config=subgraph_config, module_loader=self.module_loader
        )
        # subgraph_instance.setup(self.pipeline_config)
        return subgraph_instance.build(subgraph_config)

    @abstractmethod
    @lru_cache()
    def dataset_load(
        self,
        name,
        version="latest",
        datastore=None,
        path_on_datastore=None,
        description=None,
    ):
        """Loads a dataset by either id or name. If the workspace does not contain this dataset and path is given, create the dataset.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")
            datastore (str): datastore for registering the dataset as <name>
            path_on_datastore (str): path for registering the dataset as <name>
            description (str): description of dataset to register
        """
        pass

    @abstractmethod
    def connect(self):
        """Connect to the AML workspace using internal config"""
        pass

    @staticmethod
    def validate_experiment_name(name):
        """
        Check whether the experiment name is valid. It's required that
        experiment names must be between 1 to 250 characters, start with
        letters or numbers. Valid characters are letters, numbers, "_",
        and the "-" character.
        """
        if len(name) < 1 or len(name) > 250:
            raise ValueError("Experiment names must be between 1 to 250 characters!")
        if not re.match("^[a-zA-Z0-9]$", name[0]):
            raise ValueError("Experiment names must start with letters or numbers!")
        if not re.match("^[a-zA-Z0-9_-]*$", name):
            raise ValueError(
                "Valid experiment names must only contain letters, numbers, underscore and dash!"
            )
        return True

    @staticmethod
    def validate_connection_attribute(connection_attribute, connection_attribute_type):
        """
        Function that will verify whether the given connection_attribute (Workspace name, RG name, or Subscription Id) matches the constraints on length, allowed characters, etc...
        """

        # initialize min_length and max_length
        min_length = 0
        max_length = sys.maxsize

        # populate the constraints based on connection_attribute_type
        if connection_attribute_type == "workspace":
            min_length = 3
            max_length = 33
            # from https://docs.microsoft.com/en-us/azure/templates/microsoft.machinelearningservices/workspaces?tabs=bicep#workspaces
            connection_attribute_type_constraints = (
                "only alphanumeric characters and hyphens"
            )
            regex_target = "^[a-zA-Z0-9-]*$"
        elif connection_attribute_type == "resource_group":
            # from https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/resource-name-rules#microsoftresources
            connection_attribute_type_constraints = "alphanumeric, underscore, parentheses, hyphen, period (except at end), and Unicode characters that match the allowed characters"
            regex_target = "^[-\w\._\(\)]+[-\w_\(\)]$"
        elif connection_attribute_type == "subscription_id":
            # can't find an official link for this...
            connection_attribute_type_constraints = "V4 GUID"
            # following regex taken from: https://stackoverflow.com/questions/19989481/how-to-determine-if-a-string-is-a-valid-v4-uuid
            regex_target = "^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$"
        else:
            regex_target = ""
            connection_attribute_type_constraints = ""
            raise ValueError(
                "connection_attribute_type must be one of 'workspace', 'resource_group', or 'subscription_id'"
            )

        # check the length first
        if (len(connection_attribute) < min_length) or (
            len(connection_attribute) > max_length
        ):
            raise ValueError(
                f"{connection_attribute} must be between {min_length} and {max_length} characters."
            )

        # check whether the regex matches
        if not re.match(regex_target, connection_attribute):
            raise ValueError(
                f"Valid {connection_attribute_type} names have the following constraints: {connection_attribute_type_constraints}"
            )

        return True

    @abstractmethod
    def is_eyesoff(self) -> bool:
        """ "
        Check whether the workspace is eyes-off.
        If it lives in a non-Torus tenant, then eyes-off;
        If in Torus tenant, check whether it is in the allow-list of eyes-on Torus subscriptions.
        """
        pass

    #######################
    ### HELPER BACKEND  ###
    #######################

    @classmethod
    def _default_config(cls):
        """Builds the default config for the pipeline class"""
        config_store = ConfigStore.instance()

        config_dict = default_config_dict()
        cls._build_config(config_dict)

        config_store.store(name="default", node=config_dict)
        return OmegaConf.structured(config_dict)

    @classmethod
    def _build_config(cls, config_dict, modules_config=None):
        """Builds the entire configuration object for this graph and all subgraphs."""
        self_config_class = cls.get_config_class()
        if self_config_class:
            config_dict[self_config_class.__name__] = self_config_class

        for subgraph_key, subgraph_class in cls.required_subgraphs().items():
            subgraph_class._build_config(config_dict)

    @abstractmethod
    def _dataset_load_by_name_or_id(self, name, version):
        """Loads a dataset by either id or name.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")
        """
        pass

    @abstractmethod
    def apply_recommended_runsettings(
        self,
        module_name,
        module_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep="auto",
        synapse="auto",
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            gpu (bool): is the module using GPU?
            hdi (bool): is the module using HDI/Spark?
            windows (bool): is the module using Windows compute?
            parallel (bool): is the module using ParallelRunStep?
            mpi (bool): is the module using Mpi?
            scope (bool): is the component using scope?
            datatransfer (bool): is the component using datatransfer?
            sweep (bool): is the component using sweep?
            synapse (bool): is the component using Synapse?
            custom_runtime_arguments (dict): any additional custom args
        """
        pass

    def _parse_pipeline_tags(self):
        """Parse the tags specified in the pipeline yaml"""
        pipeline_tags = {}
        if self.config.run.tags:
            if isinstance(self.config.run.tags, str):
                try:
                    # json.load the tags string in the config
                    pipeline_tags = json.loads(self.config.run.tags)
                except ValueError:
                    log.warning(
                        f"The pipeline tags {self.config.run.tags} is not a valid json-style string."
                    )
            elif isinstance(self.config.run.tags, DictConfig):
                # NOTE: values for tags need to be str
                for key, value in self.config.run.tags.items():
                    pipeline_tags[key] = str(value)
            else:
                log.warning(
                    f"The pipeline tags {self.config.run.tags} is not a valid DictConfig or json-style string."
                )
        return pipeline_tags

    def _check_if_spec_yaml_override_is_needed(self):
        if self.config.module_loader.use_local == "":
            log.info(
                "All components are using remote copy, so override will not be executed. For components you want submission-time override of images/tags/etc., please specify them in `use_local`."
            )
            return False, None
        if not self.config.tenant_overrides.allow_override:
            log.info(
                "Spec yaml file override is not allowed. If you want to use this feature, please set `tenant_overrides.allow_override` to True in your pipeline yaml."
            )
            return False, None
        cur_tenant = self.config.aml.tenant
        for tenant in self.config.tenant_overrides.mapping.keys():
            if tenant == cur_tenant:
                log.info(
                    f"Your tenant is inconsistent with spec yaml, We will override relevant fields in your spec yaml files based on entry `{tenant}` in your pipeline yaml file."
                )
                return True, self.config.tenant_overrides.mapping[tenant]
            if self.config.run.config_dir:
                config_file_path = os.path.join(
                    self.config.run.config_dir, "aml", tenant + ".yaml"
                )
                log.info(f"config_file_path is {config_file_path}")
                if os.path.exists(config_file_path):
                    with open(config_file_path, "r") as file:
                        config = yaml.safe_load(file)
                    if config["tenant"] == cur_tenant:
                        log.info(
                            f"Your tenant is inconsistent with spec yaml, We will override relevant fields in your spec yaml files based on entry `{config_file_path}` in your pipeline yaml file."
                        )
                        return True, self.config.tenant_overrides.mapping[tenant]
        return False, None

    def _override_spec_yaml(self, spec_mapping):
        module_keys = self.module_loader.modules_manifest
        yaml_to_be_recovered = []
        env_yaml_override_is_needed = (
            spec_mapping.remove_polymer_pkg_idx
            if "remove_polymer_pkg_idx" in spec_mapping
            else False
        )
        for module_key in module_keys:
            if not self.module_loader.is_local(module_key):
                log.info(
                    f"Component {module_key} is using the remote copy. Skipping overrides."
                )
                continue
            log.info(f"Overriding for component: {module_key}.")
            module_entry = self.module_loader.modules_manifest[module_key]
            spec_path = os.path.join(
                self.module_loader.local_steps_folder, module_entry["yaml"]
            )
            (
                old_spec_path,
                old_env_file_path,
                new_env_file_path,
            ) = self._override_single_spec_yaml(
                spec_path, spec_mapping, env_yaml_override_is_needed
            )

            yaml_to_be_recovered.append(
                [old_spec_path, spec_path, old_env_file_path, new_env_file_path]
            )
        return yaml_to_be_recovered

    def _update_value_given_flattened_key(self, nested_dict, dot_key, new_val):
        log.info(f"Updating key {dot_key}")
        split_key = dot_key.split(".")
        res = nested_dict
        path = ""
        for key in split_key[:-1]:
            path += key + "."
            if not isinstance(res, dict) or key not in res:
                raise KeyError(
                    f"Key {dot_key} not in {nested_dict}. It failed at {path}."
                )
            res = res[key]
        if isinstance(res, dict) and split_key[-1] in res:
            res[split_key[-1]] = new_val
            log.info(f"The field {dot_key} has been updated to {new_val} successfully.")
        else:
            raise KeyError(f"Key {dot_key} not in {nested_dict}.")

    def _override_single_spec_yaml(
        self, spec_path, spec_mapping, env_yaml_override_is_needed
    ):
        spec_filename, spec_ext = os.path.splitext(spec_path)
        old_spec_path = (
            spec_filename + ".not_used"
        )  # remove ".yaml" extension to avoid confusion
        shutil.copyfile(
            spec_path, old_spec_path
        )  # need to recover after pipeline submission

        with open(spec_path) as file:
            spec = yaml.safe_load(file)
        for key in spec_mapping:
            match = jsonpath_ng.parse(key).find(spec)
            if match:
                try:
                    log.info(f"Find a matching field to override: {key}.")
                    original_val = match[0].value
                    log.info(
                        f"Original value is {original_val}. Looking for new value now..."
                    )
                    spec_mapping_to_use = spec_mapping[key]
                    if isinstance(original_val, str):
                        log.info(f"The field to be updated is str.")
                        if original_val in spec_mapping_to_use:
                            new_val = spec_mapping_to_use[original_val]
                            log.info(f"The new value is: {new_val}.")
                            self._update_value_given_flattened_key(spec, key, new_val)
                        else:
                            for pattern in spec_mapping_to_use:
                                if re.match(pattern, original_val):
                                    new_val = spec_mapping_to_use[pattern]
                                    log.info(f"The new pattern is: {new_val}.")
                                    self._update_value_given_flattened_key(
                                        spec,
                                        key,
                                        re.sub(pattern, new_val, original_val),
                                    )
                    elif isinstance(original_val, dict):
                        log.info("The field to be updated is dict")
                        for spec_mapping_key in spec_mapping_to_use:
                            self._update_value_given_flattened_key(
                                spec,
                                ".".join([key, spec_mapping_key]),
                                spec_mapping_to_use[spec_mapping_key],
                            )
                    else:
                        log.info(
                            f"Override for key {key} is not supported yet. Please open a [feature request](https://github.com/ai-platform-ml-platform/shrike/issues) if necessary."
                        )
                except KeyError:
                    log.info(f"Key {key} does not in file {spec_path}. Skip overrides.")
        new_env_file_path = None
        old_env_file_path = None
        if env_yaml_override_is_needed:
            spec_dirname = os.path.dirname(spec_path)
            log.info(
                f"Will remove {POLYMER_FEED} and {O365_FEED} from environment.conda."
            )
            try:
                if spec["type"] == "spark":
                    conda_dependencies_file = spec["environment"]["conda_file"]
                else:
                    conda_dependencies_file = spec["environment"]["conda"][
                        "conda_dependencies_file"
                    ]
                log.info("conda_dependencies_file exists.")
                (
                    found_index_url,
                    new_file,
                    new_env_file_path,
                    old_env_file_path,
                ) = self._remove_python_package_feed_if_exists_and_save_new(
                    spec_dirname, conda_dependencies_file, [POLYMER_FEED, O365_FEED]
                )
                if found_index_url:
                    if spec["type"] == "spark":
                        spec["environment"]["conda_file"] = new_file
                    else:
                        spec["environment"]["conda"][
                            "conda_dependencies_file"
                        ] = new_file
            except KeyError:
                # conda_dependencies_file does not exist
                pass
            try:
                if spec["type"] == "spark":
                    conda_dependencies = spec["conda_dependencies"]["dependencies"]
                else:
                    conda_dependencies = spec["environment"]["conda"][
                        "conda_dependencies"
                    ]["dependencies"]
                log.info("conda_dependencies_file exists.")
                for idx, dependency in enumerate(conda_dependencies):
                    if isinstance(dependency, dict) and "pip" in dependency:
                        pip_dependencies = dependency["pip"]
                        if POLYMER_FEED in pip_dependencies:
                            pip_dependencies.remove(POLYMER_FEED)
                        if O365_FEED in pip_dependencies:
                            pip_dependencies.remove(O365_FEED)
                        dependency["pip"] = pip_dependencies
                        conda_dependencies[idx] = dependency
                if spec["type"] == "spark":
                    spec["conda_dependencies"]["dependencies"] = conda_dependencies
                else:
                    spec["environment"]["conda"]["conda_dependencies"][
                        "dependencies"
                    ] = conda_dependencies
            except KeyError:
                # conda_dependencies does not exist
                pass
            try:
                pip_requirements_file = spec["environment"]["conda"][
                    "pip_requirements_file"
                ]
                log.info("pip_requirements_file exists.")
                (
                    found_index_url,
                    new_file,
                    new_env_file_path,
                    old_env_file_path,
                ) = self._remove_python_package_feed_if_exists_and_save_new(
                    spec_dirname, pip_requirements_file, [POLYMER_FEED, O365_FEED]
                )
                if found_index_url:
                    spec["environment"]["conda"]["pip_requirements_file"] = new_file
            except KeyError:
                # pip_requirements_file does not exist
                pass

        with open(spec_path, "w") as file:
            yaml.safe_dump(spec, file)
        return old_spec_path, old_env_file_path, new_env_file_path

    def _remove_python_package_feed_if_exists_and_save_new(
        self, spec_dirname, file, package_feeds
    ):
        found_index_url = False
        new_file = ""
        new_file_path = ""
        old_file_path = ""
        with open(os.path.join(spec_dirname, file), "r") as f:
            lines = f.readlines()
        for line in lines:
            for feed in package_feeds:
                if feed in line:
                    found_index_url = True
                    lines.remove(line)
                    break
        if found_index_url:
            file_name, file_ext = os.path.splitext(file)
            new_file = file_name + "_" + self.config.aml.tenant + file_ext
            new_file_path = os.path.join(spec_dirname, new_file)
            with open(new_file_path, "w") as f:
                f.writelines(lines)
            old_file_path = os.path.join(
                spec_dirname, os.path.splitext(file)[0] + ".not_used"
            )
            shutil.move(os.path.join(spec_dirname, file), old_file_path)
        return found_index_url, new_file, new_file_path, old_file_path

    def _recover_spec_yaml(self, spec_file_pairs, keep_modified_files):
        log.info(
            f"Reverting changes to spec yaml files. Keeping modified spec yaml files: {keep_modified_files}."
        )
        for (
            old_spec_path,
            new_spec_path,
            old_env_file_path,
            new_env_file_path,
        ) in spec_file_pairs:
            # Question: do we want to keep a copy of the modified files?
            if keep_modified_files:
                filename, ext = os.path.splitext(new_spec_path)
                shutil.move(
                    new_spec_path, filename + "_" + self.config.aml.tenant + ext
                )
                if os.path.exists(filename + ".additional_includes"):
                    shutil.copyfile(
                        filename + ".additional_includes",
                        filename
                        + "_"
                        + self.config.aml.tenant
                        + ".additional_includes",
                    )
            else:
                if new_env_file_path:
                    os.remove(new_env_file_path)
            shutil.move(old_spec_path, new_spec_path)
            if old_env_file_path:
                shutil.move(
                    old_env_file_path, os.path.splitext(old_env_file_path)[0] + ".yaml"
                )

    def _recover_tenant_overrides(
        self, override, yaml_to_be_recovered, keep_modified_files
    ):
        if override and yaml_to_be_recovered:
            try:
                self._recover_spec_yaml(yaml_to_be_recovered, keep_modified_files)
            except BaseException as e:
                log.error(f"An error occurred, recovery is not successful: {e}")

    @lru_cache()
    def _get_component_name_from_instance(self, component_instance):
        """
        We need to have this `_get_component_name_from_instance()` to get
        component name for `apply_smart_runsettings()`, and can't simply use
        `component_name = component_instance.name`. Otherwise
        `apply_smart_runsettings()` might work incorrectly for local
        components. See more detailed explanation and examples here:
        https://github.com/ai-platform-ml-platform/shrike/issues/411
        """
        component_manifest_list = self.config.modules.manifest
        component_name = self._get_component_name_helper(component_instance)

        for component_manifest_entry in component_manifest_list:
            try:
                if component_manifest_entry["name"].lower() == component_name.lower():
                    return component_manifest_entry["key"] or component_name
                if "namespace" in component_manifest_entry:
                    component_entry_name = (
                        component_manifest_entry["namespace"]
                        + "://"
                        + component_manifest_entry["name"]
                    )
                    if component_entry_name.lower() == component_name.lower():
                        return component_manifest_entry["key"] or component_name
            except ValueError:
                pass
        raise ValueError(
            f"Could not find component matching {component_name}. Please check your spelling."
        )

    def _check_module_runsettings_consistency(self, module_key, module_instance):
        """Verifies if entry at module_key matches the module instance description"""
        (
            module_manifest_entry,
            _,
            _,
        ) = self.module_loader.get_module_manifest_entry(
            module_key, modules_manifest=self.required_modules()
        )

        if "name" in module_manifest_entry:
            if ("source" in module_manifest_entry) and (module_manifest_entry.source):
                # This indicates that user is using the deprecated 'source' config in their yaml file, since
                # the default value of 'source' config has been override. Throw out warning and ask users to remove it.
                log.warning(
                    "!!IMPORTANT: Key word `source` is deprecated for module artifacts. Please remove it."
                )
            module_instance_name = self._get_component_name_helper(module_instance)

            if module_manifest_entry["name"].lower() == module_instance_name.lower():
                return
            if "namespace" in module_manifest_entry:
                module_entry_name = (
                    module_manifest_entry["namespace"]
                    + "://"
                    + module_manifest_entry["name"]
                )
                if module_entry_name.lower() == module_instance_name.lower():
                    return
            raise Exception(
                f"During build() of graph class {self.__class__.__name__}, call to self.apply_recommended_runsettings() is wrong: key used as first argument ('{module_key}') maps to a module reference {module_manifest_entry} which name is different from the module instance provided as 2nd argument (name={module_instance_name}), did you use the wrong module key as first argument?"
            )

    def _validate_tags(self, tags: dict):
        validated_tags = {}
        if isinstance(tags, dict):
            # NOTE: values for tags need to be str
            for key, value in tags.items():
                if not isinstance(key, str):
                    log.warning(f"The key for pipeline tag {key} is not string")
                else:
                    validated_tags[key] = str(value)
        else:
            log.warning(f"The pipeline tags {tags} value type is not a dict")
        return validated_tags

    ################
    ### MAIN/RUN ###
    ################

    @abstractmethod
    def _build_pipeline(self):
        pass

    @abstractmethod
    def _submit_pipeline(self, pipeline):
        pass

    def build_and_submit_new_pipeline(self):
        """Build the pipeline and submit it."""
        pipeline = self._build_pipeline()
        return self._submit_pipeline(pipeline)

    def run(self, pipeline=None) -> None:
        """Run pipeline using arguments.

        Args:
            pipeline: If set to `None`, the pipeline helper will take care of
              building a pipeline via it's `build` functions, otherwise the
              pipeline passed here will be used.
        """
        # set logging level
        if self.config.run.verbose:
            logging.getLogger().setLevel(logging.DEBUG)
        elif self.config.run.log_error_only:
            logging.getLogger().setLevel(logging.ERROR)

        # Log the telemetry information in the Azure Application Insights
        try:
            telemetry_logger = TelemetryLogger(
                enable_telemetry=not self.config.run.disable_telemetry
            )
            telemetry_logger.log_trace(
                message=f"shrike.pipeline=={__version__}",
                properties={"custom_dimensions": {"configuration": str(self.config)}},
            )
        except Exception as ex:
            log.debug(
                f"Sending trace log messages to application insight is not successful. The exception message: {ex}"
            )

        # Check whether the experiment name is valid
        self.validate_experiment_name(self.config.run.experiment_name)

        # Check whether the workspace name, resource group name, and subscription id are valid
        self.validate_connection_attribute(
            self.config.aml.workspace_name, "workspace"
        )  # self.config.aml.workspace_name
        self.validate_connection_attribute(
            self.config.aml.resource_group, "resource_group"
        )
        self.validate_connection_attribute(
            self.config.aml.subscription_id, "subscription_id"
        )

        self.repository_info = get_repo_info()
        log.info(f"Running from repository: {self.repository_info}")
        # TODO: log SDK version

        self.connect()

        pipeline_job = None
        if self.config.run.resume:
            if pipeline is not None:
                raise Exception(
                    "Did not expect a custom pipeline instance to be given when resuming a pipeline run ."
                    "Either pass `pipeline=None` or set `self.config.run.resume=False`."
                )
            if not self.config.run.pipeline_run_id:
                raise Exception(
                    "To be able to use --resume you need to provide both --experiment-name and --run-id."
                )

            log.info(f"Resuming Experiment {self.config.run.experiment_name}...")
            # experiment = Experiment(self.ml_client(), self.config.run.experiment_name)
            log.info(f"Resuming PipelineRun {self.config.run.pipeline_run_id}...")
            # pipeline_run is of the class "azureml.pipeline.core.PipelineRun"
            # pipeline_run = PipelineRun(experiment, self.config.run.pipeline_run_id)
            log.error(f"Resuming pipeline run is not supported yet.")
            raise NotImplementedError("Resuming pipeline run is not supported yet.")
        else:
            keep_modified_files, override = False, False
            yaml_to_be_recovered = []

            if self.config.tenant_overrides.allow_override:
                log.info("Check if tenant is consistent with spec yaml")
                override, mapping = self._check_if_spec_yaml_override_is_needed()
                if override:
                    try:
                        tenant = self.config.aml.tenant
                        log.info(
                            f"Performing spec yaml override to adapt to tenant: {tenant}."
                        )
                        yaml_to_be_recovered = self._override_spec_yaml(mapping)
                        keep_modified_files = (
                            self.config.tenant_overrides.keep_modified_files
                        )
                    except BaseException as e:
                        log.error(f"An error occurred, override is not successful: {e}")
                        raise e
            try:
                if pipeline is None:
                    pipeline_job = self.build_and_submit_new_pipeline()
                else:
                    pipeline_job = self._submit_pipeline(pipeline=pipeline)
            except BaseException as e:
                log.error(f"An error {e} occurred during pipeline submission.")
                if override:
                    log.info("Now trying to recover overrides.")
                self._recover_tenant_overrides(override, yaml_to_be_recovered, False)
                raise
            else:
                self._recover_tenant_overrides(
                    override, yaml_to_be_recovered, keep_modified_files
                )

        if not pipeline_job:
            # not submitting code, exit now
            return
        # launch the pipeline execution
        log.info(f"Pipeline Run Id: {pipeline_job.id}")
        log.info(
            f"""
#################################
#################################
#################################

Follow link below to access your pipeline run directly:
-------------------------------------------------------

{self._get_run_url(pipeline_job)}

#################################
#################################
#################################
        """
        )

        if self.config.run.canary:
            log.info(
                "*** CANARY MODE ***\n----------------------------------------------------------"
            )
            self._wait_for_job_completion(pipeline_job)

            # azureml.pipeline.core.PipelineRun.get_status(): ["Running", "Finished", "Failed"]
            # azureml.core.run.get_status(): ["Running", "Completed", "Failed"]
            if self._get_job_status(pipeline_job) in ["Finished", "Completed"]:
                log.info("*** PIPELINE FINISHED, TESTING WITH canary() METHOD ***")
                self.canary(self.config, pipeline_job.experiment, pipeline_job)
                log.info("OK")
                step_runs = pipeline_job.get_steps()
                for step_run in step_runs:
                    pass
                try:
                    datasets = []
                    for output in step_run.get_details()["outputDatasets"]:
                        dataset = output["dataset"]
                        dataset_info = dataset._dataflow._steps[0].arguments._pod[
                            "datastores"
                        ][0]
                        print(f"Output dataset {dataset_info}.")
                        datasets.append(
                            Data(
                                datastore_name=dataset_info["datastoreName"],
                                path=dataset_info["path"],
                            )
                        )
                    return datasets
                except:
                    print("no output")
            elif self._get_job_status(pipeline_job) == "Failed":
                log.info("*** PIPELINE FAILED ***")
                raise Exception("Pipeline failed.")
            else:
                log.info("*** PIPELINE STATUS {} UNKNOWN ***")
                raise Exception("Pipeline status is unknown.")

        else:
            if not self.config.run.silent:
                webbrowser.open(url=self._get_run_url(pipeline_job))

            # This will wait for the completion of the pipeline execution
            # and show the full logs in the meantime
            if self.config.run.resume or self.config.run.wait:
                log.info(
                    "Below are the raw debug logs from your pipeline execution:\n----------------------------------------------------------"
                )
                pipeline_job.wait_for_completion(show_output=True)

    @lru_cache()
    def _get_datastore_name(self, module_instance, compliant, component_name=None):
        if compliant:
            component_name = component_name or self._get_component_name_from_instance(
                module_instance
            )
            if self.module_loader.is_local(component_name):
                datastore_name = self.config.compute.dc_datastore
                if not datastore_name:
                    if self.is_eyesoff():
                        raise ShrikeUserErrorException(
                            "Please specify `dc_datastore` in your compute yaml so that local components can write to it."
                        )
                    else:
                        log.warning(
                            "We recommend specifying `dc_datastore` in your compute yaml so that local components can write to it. Using `compliant_datastore` for now."
                        )
                        datastore_name = self.config.compute.compliant_datastore
            else:
                datastore_name = self.config.compute.compliant_datastore
        else:
            datastore_name = self.config.compute.noncompliant_datastore
        return datastore_name

    @abstractmethod
    def _get_pipeline_run(self, run_id: str):
        pass

    @abstractmethod
    def _get_run_url(self, pipeline_run) -> str:
        pass

    @abstractmethod
    def _set_all_inputs_to(self, module_instance, input_mode):
        pass

    @abstractmethod
    def _set_all_outputs_to(
        self, module_instance, output_mode, compliant=True, datastore_name=None
    ):
        pass

    @abstractmethod
    def _get_component_name_helper(self, component_instance):
        pass

    @classmethod
    def main(cls):
        """Pipeline helper main function, parses arguments and run pipeline."""
        config_dict = cls._default_config()

        @hydra.main(config_name="default")
        def hydra_run(cfg: DictConfig):
            # merge cli config with default config
            cfg = OmegaConf.merge(config_dict, cfg)

            arg_parser = argparse.ArgumentParser()
            arg_parser.add_argument("--config-dir")
            args, _ = arg_parser.parse_known_args()
            cfg.run.config_dir = os.path.join(
                HydraConfig.get().runtime.cwd, args.config_dir
            )

            log.info("*** CONFIGURATION ***")
            log.info(OmegaConf.to_yaml(cfg))

            # create class instance
            main_instance = cls(cfg)

            # run
            main_instance.run()

        hydra_run()

        return cls.BUILT_PIPELINE  # return so we can have some unit tests done



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Pipeline helper class to create pipelines loading modules from a flexible manifest.
"""
from abc import ABC
import json
import logging
import yaml
from typing import Any, Callable, Dict, List, Optional, Set, Union

from toposort import toposort_flatten
import datetime
import time
from dataclasses import dataclass
from functools import lru_cache, wraps

from shrike.pipeline.pipeline_helper_base import AMLPipelineHelperBase

try:
    from omegaconf import DictConfig, OmegaConf, MISSING
    from flatten_dict import flatten

    from azure.ai.ml import Input
    from azure.ai.ml.entities._builders import BaseNode
    from azure.ai.ml.entities import (
        # Component types
        Spark,
        Parallel,
        Sweep,
        Command,
        # Concepts types
        Component,
        PipelineJob,
        Job,
        PipelineComponent,
        Pipeline,
        BatchEndpoint,
        PipelineComponentBatchDeployment,
        JobSchedule,
        RecurrenceTrigger,
        RecurrencePattern,
        CronTrigger,
    )

    from azure.ai.ml.exceptions import ValidationException, UserErrorException
    from azure.ai.ml._internal.entities import (
        Command as InternalCommand,
        Distributed,
        Parallel as InternalParallel,
        Scope,
        DataTransfer,
        Ae365exepool,
        HDInsight,
        Starlite,
        Hemera,
    )
    from azure.ai.ml.entities._job.sweep.early_termination_policy import (
        BanditPolicy,
        MedianStoppingPolicy,
        TruncationSelectionPolicy,
    )
    from azure.ai.ml.entities._job.pipeline._io import (
        NodeInput,
        NodeOutput,
        InputOutputBase,
    )
    from azure.ai.ml.entities._job.job_resource_configuration import (
        JobResourceConfiguration,
    )
    from azure.ai.ml.entities._job.parallel.retry_settings import RetrySettings
    from azure.ai.ml.entities._job.distribution import DistributionConfiguration
except ImportError as error:
    raise ImportError(
        f"{error.msg}. Please install v2 dependency `pip install azure-ai-ml`."
    )

from shrike import __version__
from shrike.pipeline.aml_connect import (
    azureml_connect,
    current_ml_client,
    MLClientNotInitializedError,
)
from shrike.pipeline.module_helper import AMLModuleLoader
from shrike.pipeline.pipeline_config import (
    default_config_dict,
    HDI_DEFAULT_CONF,
)
from shrike._core import is_eyesoff_helper

log = logging.getLogger(__name__)


@dataclass
class Data:
    datastore_name: str
    path: str


class AMLPipelineHelper(AMLPipelineHelperBase, ABC):
    """Helper class for building v2 pipelines"""

    BUILT_PIPELINE = None  # the hydra run decorator doesn't allow for return, we're using this variable instead (hack)

    def __init__(self, config, module_loader=None):
        """Constructs the pipeline helper.

        Args:
            config (DictConfig): config for this object
            module_loader (AMLModuleLoader): which module loader to (re)use
        """
        if module_loader is None:
            log.info(
                f"Creating instance of AMLModuleLoader for {self.__class__.__name__}"
            )
            module_loader = AMLModuleLoader(config)
        super(AMLPipelineHelper, self).__init__(
            config=config, module_loader=module_loader
        )

    def _get_job_status(self, job):
        return self.ml_client().jobs.get(job.name).status

    def _wait_for_job_completion(self, job):
        return self.ml_client().jobs.stream(job.name)

    ##################################
    ### USER FACING HELPER METHODS ###
    ##################################

    @lru_cache()
    def component_load(self, component_key) -> Callable[..., "Component"]:
        """Loads one component from the manifest"""
        try:
            component_func = self.module_loader.load_module(
                component_key, self.required_modules()
            )
        except ValidationException as e:
            raise RuntimeError(
                e.message
                + f"The type of {component_key} is not supported in v2, need to convert it to v2 format."
            )
        self._check_component_supported(component_func)

        @wraps(component_func)
        def wrapper(*args, **kwargs):
            component_step = component_func(*args, **kwargs)
            log.info(
                f"Now try applying the smart runsettting as component post load operation after component initialize: "
            )
            self.apply_smart_runsettings(component_step)
            log.info(f"Successfully applied the smart runsettting. ")
            return component_step

        return wrapper

    @lru_cache()
    def module_load(self, module_key):
        """Loads one module from the manifest"""
        try:
            module_func = self.module_loader.load_module(
                module_key, self.required_modules()
            )
        except ValidationException as e:
            raise RuntimeError(
                e.message
                + f"The type of {module_key} is not supported in v2, need to convert it to v2 format."
            )
        self._check_component_supported(module_func)

        @wraps(module_func)
        def wrapper(*args, **kwargs):
            module_step = module_func(*args, **kwargs)
            self.apply_smart_runsettings(module_step)
            return module_step

        return wrapper

    @lru_cache()
    def dataset_load(
        self,
        name,
        version="latest",
        datastore=None,
        path_on_datastore=None,
        description=None,
    ):
        """Loads a dataset by either id or name. If the workspace does not contain this dataset and path is given, create the dataset.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")
            datastore (str): datastore for registering the dataset as <name>
            path_on_datastore (str): path for registering the dataset as <name>
            description (str): description of dataset to register
        """
        try:
            dataset = self._dataset_load_by_name_or_id(name, version)
            if "v1_type" in dataset.properties:
                return Input(path=dataset.path)
            else:
                return dataset
        except Exception:
            log.info(
                f"Dataset {name} not found. Try registering from {path_on_datastore} on {datastore}..."
            )
            if not datastore or not path_on_datastore:
                raise RuntimeError(
                    f"Dataset {name} not found and no path or datastore given."
                )
            # No need to create data if it comes from path on datastore, pipeline can directly consume it.
            data = Input(
                path=f"azureml://datastores/{datastore}/paths/{path_on_datastore}"
            )
            return data

    @lru_cache()
    def ml_client(self):
        """Gets the current ml_client"""
        return current_ml_client()

    def connect(self):
        """Connect to the AML workspace using internal config"""
        # Only call azureml_connect if there is not an already a pre-existing workspace
        try:
            return self.ml_client()
        except MLClientNotInitializedError:
            return azureml_connect(
                aml_subscription_id=self.config.aml.subscription_id,
                aml_resource_group=self.config.aml.resource_group,
                aml_workspace_name=self.config.aml.workspace_name,
                aml_auth=self.config.aml.auth,
                aml_tenant=self.config.aml.tenant,
                aml_force=self.config.aml.force,
            )  # NOTE: this also stores aml workspace in internal global variable

    def apply_recommended_runsettings(
        self,
        module_name,
        module_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep=False,  # can't autodetect that
        synapse="auto",
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            gpu (bool): is the module using GPU?
            hdi (bool): is the module using HDI?
            windows (bool): is the module using Windows compute?
            parallel (bool): is the module using ParallelRunStep?
            mpi (bool): is the module using Mpi?
            scope (bool): is the component using scope?
            datatransfer (bool): is the component using datatransfer?
            sweep (bool): is the component using sweep?
            synapse (bool): is the component using Synapse?
            custom_runtime_arguments (dict): any additional custom args
        """
        # Verifies if module_name corresponds to module_instance when component's available
        if isinstance(module_instance.component, Component):
            self._check_module_runsettings_consistency(module_name, module_instance)

        # Skips if this is a PipelineComponent
        if str(module_instance.type) in ["PipelineComponent", "pipeline"]:
            log.info(f"Component {module_name} detected as PipelineComponent.")
            return

        # Auto detects runsettings
        if hdi == "auto":
            hdi = str(module_instance.type) == "HDInsightComponent"
            if hdi:
                log.info(f"Module {module_name} detected as HDI: {hdi}")

        if parallel == "auto":
            parallel = str(module_instance.type) in ["ParallelComponent", "parallel"]
            if parallel:
                log.info(f"Module {module_name} detected as PARALLEL: {parallel}")

        if mpi == "auto":
            mpi = str(module_instance.type) == "DistributedComponent"
            if mpi:
                log.info(f"Module {module_name} detected as MPI: {mpi}")

        if scope == "auto":
            scope = str(module_instance.type) == "ScopeComponent"
            if scope:
                log.info(f"Module {module_name} detected as SCOPE: {scope}")

        if sweep == "auto":
            sweep = False
            log.warning(
                "Sweep components cannot be auto-detected. Please set `sweep=True` in `apply_smart_runsettings`."
            )

        if synapse == "auto":
            synapse = (
                str(module_instance.type) == "spark"
                or str(module_instance.type) == "SparkComponent"
            )
            if synapse:
                log.info(f"Module {module_name} detected as spark (Synapse): {synapse}")

        if windows == "auto":
            if (
                str(module_instance.type) == "HDInsightComponent"
                or str(module_instance.type) == "ScopeComponent"
                or str(module_instance.type) == "DataTransferComponent"
                or str(module_instance.type) == "sweep"
            ):
                # HDI/scope/datatransfer/sweep modules might not have that environment object
                windows = False
            else:
                os = getattr(
                    module_instance.component.environment, "os", "Linux"
                ) or getattr(module_instance.component.environment, "os_type", "Linux")
                # to fix AttributeError: 'NoneType' object has no attribute 'lower'
                windows = os and os.lower() == "windows"

                if windows:
                    log.info(f"Module {module_name} detected as WINDOWS: {windows}")

        if datatransfer == "auto":
            datatransfer = str(module_instance.type) == "DataTransferComponent"
            if datatransfer:
                log.info(
                    f"Module {module_name} detected as DATATRANSFER: {datatransfer}"
                )

        if parallel:
            self._apply_parallel_runsettings(
                module_name,
                module_instance,
                windows=windows,
                gpu=gpu,
                **custom_runtime_arguments,
            )
            return

        if sweep:
            self._apply_sweep_runsettings(
                module_name,
                module_instance,
                windows=windows,
                gpu=gpu,
                **custom_runtime_arguments,
            )
            return

        if windows:
            self._apply_windows_runsettings(
                module_name,
                module_instance,
                mpi=mpi,
                **custom_runtime_arguments,
            )
            return

        if hdi:
            self._apply_hdi_runsettings(
                module_name,
                module_instance,
                **custom_runtime_arguments,
            )
            return

        if scope:
            self._apply_scope_runsettings(
                module_name, module_instance, **custom_runtime_arguments
            )
            return

        if datatransfer:
            self._apply_datatransfer_runsettings(
                module_name, module_instance, **custom_runtime_arguments
            )
            return

        if synapse:
            self._apply_synapse_runsettings(
                module_name, module_instance, **custom_runtime_arguments
            )
            return

        self._apply_linux_runsettings(
            module_name,
            module_instance,
            mpi=mpi,
            gpu=gpu,
            **custom_runtime_arguments,
        )

    def apply_smart_runsettings(
        self,
        component_instance,
        gpu=False,  # can't autodetect that
        hdi="auto",
        windows="auto",
        parallel="auto",
        mpi="auto",
        scope="auto",
        datatransfer="auto",
        sweep=False,
        synapse="auto",
        **custom_runtime_arguments,
    ):
        """Applies regular settings for a given component.

        Args:
            component_instance (Component): the AML component we need to add settings to
            gpu (bool): is the component using GPU?
            hdi (bool): is the component using HDI?
            windows (bool): is the component using Windows compute?
            parallel (bool): is the component using ParallelRunStep?
            mpi (bool): is the component using Mpi?
            scope (bool): is the component using scope?
            datatransfer (bool): is the component using datatransfer?
            sweep (bool): is the component using sweep?
            synapse (bool): is the component using Synapse?
            custom_runtime_arguments (dict): any additional custom args
        """
        # infer component_name
        try:
            component_name = self._get_component_name_from_instance(component_instance)
        except Exception as e:
            # seeing error for PipelineComponent: descriptor '__str__' of 'object' object needs an argument
            component_name = self._get_component_name_helper(component_instance)
        self.apply_recommended_runsettings(
            component_name,
            component_instance,
            gpu,
            hdi,
            windows,
            parallel,
            mpi,
            scope,
            datatransfer,
            sweep,
            synapse,
            **custom_runtime_arguments,
        )

    def is_eyesoff(self) -> bool:
        """ "
        Check whether the workspace is eyes-off.
        If it lives in a non-Torus tenant, then eyes-off;
        If in Torus tenant, check whether it is in the allow-list of eyes-on Torus subscriptions.
        """
        ml_client = self.ml_client()
        # sc = SubscriptionClient(ml_client._credential)
        # for t in sc.tenants.list():
        #     pass
        tenant_id = ml_client._tenant_id
        subscription_id = ml_client.subscription_id
        return is_eyesoff_helper(tenant_id, subscription_id)

    ################
    ### MAIN/RUN ###
    ################

    def _build_pipeline(self) -> PipelineJob:
        log.info(f"Building Pipeline [{self.__class__.__name__}]...")
        pipeline_function = self.build(self.config)

        # publishing needs a Pipeline, not PipelineJob
        if self.config.run.publish:
            log.warning(
                "EXPERIMENTAL: publishing a pipeline using sdk v2 in shrike is experimental, do not take prod dependency on this."
            )
            _component = self.ml_client().components.create_or_update(pipeline_function)
            self._publish_to_endpoint(_component)

        log.info("Creating Pipeline Instance...")
        pipeline = self.pipeline_instance(pipeline_function, self.config)

        if not self.config.run.skip_update_dc and self.is_eyesoff():
            log.info("Overriding compute target if upstream steps use DC...")
            pipeline = self._update_dc_configs(pipeline)

        if not self.config.run.skip_validation:
            log.info("Validating...")
            self.ml_client().jobs.validate(pipeline)

        if self.config.run.display_name:
            pipeline.display_name = self.config.run.display_name
        if self.config.run.experiment_name:
            pipeline.experiment_name = self.config.run.experiment_name
        if self.config.run.regenerate_outputs:
            pipeline.settings.force_rerun = True
        if self.config.run.continue_on_failure:
            pipeline.settings.continue_on_step_failure = True

        if self.config.run.export:
            raise RuntimeError("Export pipeline is not supported in v2 yet.")

        if self.config.run.create_schedule:
            self._create_schedule(pipeline)

        return pipeline

    def _publish_to_endpoint(self, pipeline: Pipeline):
        endpoint_name = self.config.run.endpoint_name or self.config.run.experiment_name
        endpoint_description = (
            self.config.run.endpoint_description
            or self.config.run.experiment_description
        )
        endpoint = BatchEndpoint(name=endpoint_name, description=endpoint_description)
        self.ml_client().batch_endpoints.begin_create_or_update(endpoint)
        log.info(f"Created BatchEndpoint endpoint {endpoint.id}")

        # loop on endpoint.state until it is in "Ready" state or timeout 5 minutes
        log.info("Waiting (timeout=5mins) for endpoint to be Succeeded.")
        timeout = 300
        while timeout > 0:
            endpoint = self.ml_client().batch_endpoints.get(endpoint_name)
            if endpoint.provisioning_state == "Succeeded":
                break
            else:
                log.info(
                    f"Endpoint is not ready yet ({endpoint.provisioning_state}), waiting 1 second..."
                )
            timeout -= 1
            time.sleep(1)

        deployment_id = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
        deployment_name = f"batch-deployment-{deployment_id}"
        deployment = PipelineComponentBatchDeployment(
            name=deployment_name,
            description="EXPERIMENTAL",
            endpoint_name=endpoint.name,
            component=pipeline,
        )

        deployment = self.ml_client().batch_deployments.begin_create_or_update(
            deployment
        )
        log.info(f"Created PipelineComponentBatchDeployment {deployment_name}")

    def _submit_pipeline(self, pipeline: PipelineJob) -> Optional[Job]:
        """Publish and submit the given pipeline.

        Args:
            pipeline: pipeline to be submitted
        """
        if self.config.run.submit:
            pipeline_tags = self._create_pipeline_tags()
            log.info(f"Submitting Experiment... [tags={pipeline_tags}]")
            ml_client = self.ml_client()

            pipeline_job = ml_client.jobs.create_or_update(
                job=pipeline,
                experiment_name=self.config.run.experiment_name,
                description=self.config.run.experiment_description,
                tags=pipeline_tags,
                compute=self.config.compute.default_compute_target,
                skip_validation=self.config.run.skip_validation,
            )

            return pipeline_job

        else:
            log.info(
                "Exiting now, if you want to submit please override run.submit=True"
            )
            self.__class__.BUILT_PIPELINE = (
                pipeline  # return so we can have some unit tests done
            )
            return

    def _create_schedule(self, pipeline_job: PipelineJob):
        """Create a schedule for the given pipeline.

        Args:
            pipeline_job: pipeline to be scheduled
        """
        if self.config.run.schedule_name is None:
            raise ValueError("Schedule name must be provided.")

        if self.config.schedule_trigger.cron_expression and (
            self.config.schedule_trigger.recurrence_frequency
            or self.config.schedule_trigger.recurrence_interval
        ):
            raise ValueError(
                "When setting up a schedule, both cron_expression and recurrence_frequency/recurrence_interval are provided, please only provide one of them."
            )

        if self.config.schedule_trigger.cron_expression is None and (
            self.config.schedule_trigger.recurrence_frequency is None
            or self.config.schedule_trigger.recurrence_interval is None
        ):
            raise ValueError(
                "When setting up an schedule, either cron_expression or recurrence_frequency and recurrence_interval must be provided."
            )

        if self.config.schedule_trigger.cron_expression:
            schedule_trigger = CronTrigger(
                expression=self.config.schedule_trigger.cron_expression,
                start_time=self.config.schedule_trigger.start_time,
                end_time=self.config.schedule_trigger.end_time,
                time_zone=self.config.schedule_trigger.time_zone,
            )
        else:
            recurrence_pattern = None
            if self.config.schedule_trigger.recurrence_pattern:
                pattern_params = {
                    key: OmegaConf.to_container(value)
                    for (
                        key,
                        value,
                    ) in self.config.schedule_trigger.recurrence_pattern.items()
                    if value is not None
                }
                recurrence_pattern = RecurrencePattern(**pattern_params)

            schedule_trigger = RecurrenceTrigger(
                frequency=self.config.schedule_trigger.recurrence_frequency,
                interval=self.config.schedule_trigger.recurrence_interval,
                schedule=recurrence_pattern,
                start_time=self.config.schedule_trigger.start_time,
                end_time=self.config.schedule_trigger.end_time,
                time_zone=self.config.schedule_trigger.time_zone,
            )

        schedule_tags = self._create_pipeline_tags()
        schedule = JobSchedule(
            name=self.config.run.schedule_name,
            description=self.config.run.schedule_description,
            trigger=schedule_trigger,
            create_job=pipeline_job,
            tags=schedule_tags,
        )

        log.info(
            f"Creating Schedule for PipelineJob... [name={schedule.name}][tags={schedule_tags}][trigger={schedule_trigger}]"
        )
        self.ml_client().schedules.begin_create_or_update(schedule)

    def _get_pipeline_run(self, run_id):
        return self.ml_client().jobs.get(name=run_id)

    def _get_run_url(self, pipeline_run):
        return pipeline_run.studio_url

    def _dataset_load_by_name_or_id(self, name, version):
        """Loads a dataset by either id or name.

        Args:
            name (str): name or uuid of dataset to load
            version (str): if loading by name, used to specify version (default "latest")

        NOTE: in AzureML SDK there are 2 different methods for loading dataset
        one for id, one for name. This method just wraps them up in one."""
        # test if given name is a uuid

        # TODO(1989877): Support get data by id
        # hack here: dpv2 use tag for "latest" instead of version
        if version == "latest":
            log.info(f"Getting a dataset [name={name}] with latest label...")

            return self.ml_client().data.get(name=name, label=version)
        else:
            log.info(f"Getting a dataset [name={name}, version={version}]...")

            return self.ml_client().data.get(name=name, version=version)

    def _create_pipeline_tags(self) -> Dict[str, str]:
        pipeline_tags = self._parse_pipeline_tags()
        pipeline_tags.update({"shrike": __version__})
        pipeline_tags.update(self.repository_info)
        pipeline_tags.update(self._validate_tags(self.extra_tags()))
        return pipeline_tags

    @classmethod
    def _check_node_type(cls, node, supported_types):
        if not isinstance(node, supported_types):
            raise TypeError(
                f"Only {supported_types} is supported, got {type(node)} for node instead."
            )

    @staticmethod
    def _check_component_supported(component: Union[Component, BaseNode]):
        """Check if given component/node is supported in v2."""
        # Some component need to change to v2 format, eg: SweepComponent, SparkComponent

        if component.type == "SweepComponent":
            raise RuntimeError(
                f"{component.name} with type {component.type} is not supported in v2, need to convert it to v2 format."
            )

        supported_legacy_component_types = [
            "CommandComponent",
            "DataTransferComponent",
            "DistributedComponent",
            "HDInsightComponent",
            "ParallelComponent",
            "ScopeComponent",
            "StarliteComponent",
            "PipelineComponent",
            "HemeraComponent",
            "AE365ExePoolComponent",
            "IntellectualPropertyProtectedComponent",
        ]
        # for other v2 types: eg: automl, import, need to check if need to introduce apply_xxx_runsettings for them.
        supported_new_component_types = [
            "command",
            "sweep",
            "parallel",
            "pipeline",
            "spark",
        ]
        supported_component_types = (
            supported_legacy_component_types + supported_new_component_types
        )
        if component.type not in supported_component_types:
            raise RuntimeError(
                f"{component.name} with type {component.type} is not supported currently, "
                f"all supported types: {supported_component_types}"
            )

    def _set_all_inputs_to(self, module_instance, input_mode):
        """Sets all module inputs to a given intput mode"""
        input_names = []
        for a, val in module_instance.inputs.items():
            if isinstance(val, NodeInput):
                if isinstance(val._data, (int, bool, float, str)) or val._data is None:
                    # For primitive input, no need to set mode
                    continue
                elif isinstance(val._data, (Input, InputOutputBase)):
                    # For data input or data binding, set mode
                    input_names.append(a)
                else:
                    # Need to check if input's valid, if valid, need to support set mode.
                    raise RuntimeError(
                        f"Not supported input {a} for {self._get_component_name_from_instance(module_instance)} with value {type(val._data)}"
                    )

        for input_key in input_names:
            input_instance = getattr(module_instance.inputs, input_key)
            input_instance.mode = input_mode
            log.info(f"Configured input {input_key} to use mode {input_mode}")

    def _set_all_outputs_to(
        self, module_instance, output_mode=None, compliant=True, datastore_name=None
    ):
        """Sets all module outputs to a given output mode"""
        output_names = [
            a
            for a, val in module_instance.outputs.items()
            if isinstance(val, NodeOutput)  # and isinstance(val._data, Output)
        ]
        if not datastore_name:
            datastore_name = self._get_datastore_name(
                module_instance,
                compliant,
                self._get_component_name_from_instance(module_instance),
            )
        for output_key in output_names:
            output_instance = getattr(module_instance.outputs, output_key)

            if output_mode is not None:
                output_instance.mode = output_mode
            datastore_prefix = f"azureml://datastores/{datastore_name}/"
            output_instance.path = (
                datastore_prefix + "paths/azureml/${{name}}/outputs/${{output_name}}"
            )
            log.info(
                f"Configured output {output_key} to use mode {output_mode} and datastore {datastore_name}"
            )

    def _get_process_count_per_node(
        self, process_count_per_node, process_count_per_instance
    ):
        if not process_count_per_instance and process_count_per_node:
            log.warning(
                "`process_count_per_node` is deprecated. Please use `process_count_per_instance`."
            )
            return process_count_per_node
        else:
            return process_count_per_instance

    def _get_parallel_process_count_per_node(self):
        if (
            "parallel_process_count_per_node" in self.config.compute
            and "parallel_process_count_per_instance" not in self.config.compute
        ):
            parallel_process_count_per_instance = (
                self.config.compute.parallel_process_count_per_node
            )
            log.warning(
                "`parallel_process_count_per_node` is deprecated. Please use `parallel_process_count_per_instance` in your compute file."
            )
        else:
            parallel_process_count_per_instance = (
                self.config.compute.parallel_process_count_per_instance
                if "parallel_process_count_per_instance" in self.config.compute
                else None
            )
        return parallel_process_count_per_instance

    def _apply_windows_runsettings(
        self,
        module_name,
        module_instance,
        mpi=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        process_count_per_instance=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a windows module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            mpi (bool): is job mpi ?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): deprecated, please use `process_count_per_instance`
            process_count_per_instance (int): force process_count_per_node over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.module_loader.is_local(module_name):
            target = (
                target
                if target is not None
                else self.config.compute.windows_cpu_dc_target
            )
        else:
            target = (
                target
                if target is not None
                else self.config.compute.windows_cpu_prod_target
            )

        log.info(
            f"Using windows compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unhandled custom_runtime_arguments: {custom_runtime_arguments}"
            )
        process_count_per_instance = self._get_process_count_per_node(
            process_count_per_node, process_count_per_instance
        )

        module_instance.compute = target
        if mpi:
            node_count = node_count if node_count is not None else 1
            process_count_per_instance = (
                process_count_per_instance
                if process_count_per_instance is not None
                else 1
            )
            log.info(
                f"Using mpi with node_count={node_count} process_count_per_instance={process_count_per_instance}"
            )

            if module_instance.resources is None:
                module_instance.resources = JobResourceConfiguration()
            module_instance.resources.instance_count = node_count
            if module_instance.distribution is None:
                module_instance.distribution = DistributionConfiguration()
            module_instance.distribution.process_count_per_instance = (
                process_count_per_instance
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        else:
            self._set_all_inputs_to(
                module_instance, self.config.compute.windows_input_mode
            )

        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )
        else:
            self._set_all_outputs_to(
                module_instance,
                self.config.compute.windows_output_mode,
                datastore_name=datastore_name,
            )

    def _apply_hdi_runsettings(
        self,
        module_name,
        module_instance: HDInsight,
        target=None,
        driver_cores=None,
        driver_memory=None,
        executor_memory=None,
        executor_cores=None,
        number_executors=None,
        conf=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a HDI module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            driver_cores (int): force driver_cores over hydra conf
            driver_memory (str): force driver_memory over hydra conf
            executor_memory (int): force executor_memory over hydra conf
            executor_cores (int): force executor_cores over hydra conf
            number_executors (int): force number_executors over hydra conf
            conf (str): force conf over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """

        # check if node type is supported
        self._check_node_type(node=module_instance, supported_types=HDInsight)

        if self.module_loader.is_local(module_name):
            target = target if target is not None else self.config.compute.hdi_dc_target
            if not self.config.compute.hdi_dc_target:
                raise Exception(
                    f"Your HDI component {module_name} is using local version. Please specify hdi_dc_target"
                )
        else:
            target = (
                target if target is not None else self.config.compute.hdi_prod_target
            )
        log.info(
            f"Using HDI compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unsupported custom runtime arguments {custom_runtime_arguments}"
            )

        merged_conf = json.loads(HDI_DEFAULT_CONF)
        new_conf = (
            self.config.compute.hdi_conf if "hdi_conf" in self.config.compute else None
        )
        if conf is not None:
            new_conf = conf
        if new_conf is not None:
            if isinstance(new_conf, str):
                new_conf = json.loads(new_conf)
            elif isinstance(new_conf, DictConfig):
                new_conf = flatten(dict(new_conf), reducer="dot")
            else:
                raise ValueError(
                    "computed.hdi_conf is not a valid json string or a single tested configuration."
                )
            merged_conf.update(new_conf)

        module_instance.compute_name = target

        module_instance.driver_memory = (
            driver_memory
            if driver_memory is not None
            else self.config.compute.hdi_driver_memory
        )
        module_instance.driver_cores = (
            driver_cores
            if driver_cores is not None
            else self.config.compute.hdi_driver_cores
        )
        module_instance.executor_memory = (
            executor_memory
            if executor_memory is not None
            else self.config.compute.hdi_executor_memory
        )
        module_instance.executor_cores = (
            executor_cores
            if executor_cores is not None
            else self.config.compute.hdi_executor_cores
        )
        module_instance.number_executors = (
            number_executors
            if number_executors is not None
            else self.config.compute.hdi_number_executors
        )
        module_instance.conf = merged_conf

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

    def _apply_synapse_runsettings(
        self,
        module_name,
        module_instance: Spark,
        target=None,
        driver_cores=None,
        driver_memory=None,
        executor_memory=None,
        executor_cores=None,
        number_executors=None,
        conf=None,
        input_mode="direct",
        output_mode="direct",
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a Synapse component. Shrike will not set default runsettings except for spark_identity.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            driver_cores (int): force driver_cores over hydra conf
            driver_memory (str): force driver_memory over hydra conf
            executor_memory (int): force executor_memory over hydra conf
            executor_cores (int): force executor_cores over hydra conf
            number_executors (int): force number_executors over hydra conf
            conf (str): force conf over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """

        # check if node type is supported
        self._check_node_type(node=module_instance, supported_types=Spark)

        if self.module_loader.is_local(module_name):
            target = (
                target if target is not None else self.config.compute.synapse_dc_target
            )
            if not self.config.compute.synapse_dc_target:
                raise Exception(
                    f"Your Spark (Synapse) component {module_name} is using local version. Please specify synapse_dc_target"
                )
        else:
            target = (
                target
                if target is not None
                else self.config.compute.synapse_prod_target
            )
        log.info(
            f"Using Synapse compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unhandled custom_runtime_arguments: {custom_runtime_arguments}"
            )

        new_conf = (
            self.config.compute.synapse_conf
            if "synapse_conf" in self.config.compute
            else None
        )
        if conf is not None:
            new_conf = conf
        if new_conf is not None:
            if isinstance(new_conf, str):
                new_conf = json.loads(new_conf)
            elif isinstance(new_conf, DictConfig) or isinstance(new_conf, dict):
                new_conf = flatten(dict(new_conf), reducer="dot")
            else:
                raise TypeError(
                    f"compute.synapse_conf of type {type(new_conf)} is not a valid json string or a single tested configuration."
                )
            module_instance.conf = new_conf

        module_instance.compute = target
        module_instance.identity.type = "managed_identity"

        if "synapse_driver_memory" in self.config.compute:
            driver_memory = driver_memory or self.config.compute.synapse_driver_memory
        if driver_memory:
            module_instance.driver_memory = driver_memory
        if "synapse_driver_cores" in self.config.compute:
            driver_cores = driver_cores or self.config.compute.synapse_driver_cores
        if driver_cores:
            module_instance.driver_cores = driver_cores
        if "synapse_executor_memory" in self.config.compute:
            executor_memory = (
                executor_memory or self.config.compute.synapse_executor_memory
            )
        if executor_memory:
            module_instance.executor_memory = executor_memory
        if "synapse_number_executors" in self.config.compute:
            number_executors = (
                number_executors or self.config.compute.synapse_number_executors
            )
        if number_executors:
            module_instance.executor_instances = number_executors
        if "synapse_executor_cores" in self.config.compute:
            executor_cores = (
                executor_cores or self.config.compute.synapse_executor_cores
            )
        if executor_cores:
            module_instance.executor_cores = executor_cores

        self._set_all_inputs_to(module_instance, input_mode=input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode=output_mode, datastore_name=datastore_name
        )

    def _apply_parallel_runsettings(
        self,
        module_name,
        module_instance: Union[InternalParallel, Parallel],
        windows=False,
        gpu=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        process_count_per_instance=None,
        mini_batch_size=None,
        run_invocation_timeout=None,
        run_max_try=None,
        error_threshold=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a ParallelRunStep linux module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            windows (bool): is the module using Windows compute?
            gpu (bool): is the module using GPU compute?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): deprecated, please use `process_count_per_instance`
            process_count_per_instance (int): force process_count_per_node over hydra conf
            mini_batch_size (int): force mini_batch_size over hydra conf
            run_invocation_timeout (int): force run_invocation_timeout over hydra conf
            run_max_try (int): force run_max_try over hydra conf
            error_threshold (int): The number of file failures for the input FileDataset that should be ignored during processing.
                If the error count goes above this value, then the job will be aborted.
                Error threshold is for the entire input and not for individual mini-batches sent to run() method.
                The range is [-1, int.max]. -1 indicates ignoring all failures during processing.
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """

        # check if node type is supported
        self._check_node_type(
            node=module_instance, supported_types=(InternalParallel, Parallel)
        )

        if self.module_loader.is_local(module_name):
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    _target = self.config.compute.windows_cpu_dc_target
            else:
                if gpu:
                    _target = self.config.compute.linux_gpu_dc_target
                else:
                    _target = self.config.compute.linux_cpu_dc_target
        else:
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    _target = self.config.compute.windows_cpu_prod_target
            else:
                if gpu:
                    _target = self.config.compute.linux_gpu_prod_target
                else:
                    _target = self.config.compute.linux_cpu_prod_target

        target = target if target is not None else _target

        log.info(
            f"Using parallelrunstep compute target {target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unsupported custom runtime arguments {custom_runtime_arguments}"
            )

        module_instance.compute = target

        process_count_per_instance = self._get_process_count_per_node(
            process_count_per_node, process_count_per_instance
        )

        # Some runsettings API is different between InternalParallel and Parallel
        if isinstance(module_instance, InternalParallel):
            module_instance.resources.instance_count = (
                node_count
                if node_count is not None
                else self.config.compute.parallel_node_count
            )
            module_instance.max_concurrency_per_instance = (
                process_count_per_instance
                if process_count_per_instance is not None
                else self._get_parallel_process_count_per_node()
            )
            module_instance.mini_batch_size = str(
                mini_batch_size
                if mini_batch_size is not None
                else self.config.compute.parallel_mini_batch_size
            )

            module_instance.retry_settings.timeout = (
                run_invocation_timeout
                if run_invocation_timeout is not None
                else self.config.compute.parallel_run_invocation_timeout
            )
            module_instance.retry_settings.max_tries = (
                run_max_try
                if run_max_try is not None
                else self.config.compute.parallel_run_max_try
            )
            module_instance.error_threshold = (
                error_threshold
                if error_threshold is not None
                else self.config.compute.parallel_error_threshold
            )
        else:
            module_instance.set_resources(
                instance_count=(
                    node_count
                    if node_count is not None
                    else self.config.compute.parallel_node_count
                )
            )
            module_instance.max_concurrency_per_instance = (
                process_count_per_instance
                if process_count_per_instance is not None
                else self._get_parallel_process_count_per_node()
            )
            module_instance.mini_batch_size = str(
                mini_batch_size
                if mini_batch_size is not None
                else self.config.compute.parallel_mini_batch_size
            )

            if module_instance.retry_settings is None:
                module_instance.retry_settings = RetrySettings()
            module_instance.retry_settings.timeout = (
                run_invocation_timeout
                if run_invocation_timeout is not None
                else self.config.compute.parallel_run_invocation_timeout
            )
            module_instance.retry_settings.max_tries = (
                run_max_try
                if run_max_try is not None
                else self.config.compute.parallel_run_max_try
            )
            module_instance.error_threshold = (
                error_threshold
                if error_threshold is not None
                else self.config.compute.parallel_error_threshold
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

    def _apply_linux_runsettings(
        self,
        module_name,
        module_instance,
        mpi=False,
        gpu=False,
        target=None,
        node_count=None,
        process_count_per_node=None,
        process_count_per_instance=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for linux module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            mpi (bool): is the job mpi?
            gpu (bool): is the job using GPU?
            target (str): force target compute over hydra conf
            node_count (int): force node_count over hydra conf
            process_count_per_node (int): deprecated, please use `process_count_per_instance`
            process_count_per_instance (int): force process_count_per_node over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        if self.module_loader.is_local(module_name) and gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_gpu_dc_target
            )
            log.info(
                f"Using target {target} for local code GPU module {module_name} from pipeline class {self.__class__.__name__}"
            )
        elif not self.module_loader.is_local(module_name) and gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_gpu_prod_target
            )
            log.info(
                f"Using target {target} for registered GPU module {module_name} from pipeline class {self.__class__.__name__}"
            )
        elif self.module_loader.is_local(module_name) and not gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_cpu_dc_target
            )
            log.info(
                f"Using target {target} for local CPU module {module_name} from pipeline class {self.__class__.__name__}"
            )
        elif not self.module_loader.is_local(module_name) and not gpu:
            target = (
                target
                if target is not None
                else self.config.compute.linux_cpu_prod_target
            )
            log.info(
                f"Using target {target} for registered CPU module {module_name} from pipeline class {self.__class__.__name__}"
            )

        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unhandled custom_runtime_arguments: {custom_runtime_arguments}"
            )

        module_instance.compute = target

        process_count_per_instance = self._get_process_count_per_node(
            process_count_per_node, process_count_per_instance
        )

        if mpi:
            node_count = node_count if node_count is not None else 1
            process_count_per_instance = (
                process_count_per_instance
                if process_count_per_instance is not None
                else 1
            )
            log.info(
                f"Using mpi with node_count={node_count} process_count_per_instance={process_count_per_instance}"
            )
            if module_instance.resources is None:
                module_instance.resources = JobResourceConfiguration()
            module_instance.resources.instance_count = node_count
            if module_instance.distribution is None:
                module_instance.distribution = DistributionConfiguration()
            module_instance.distribution.process_count_per_instance = (
                process_count_per_instance
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        else:
            self._set_all_inputs_to(
                module_instance, self.config.compute.linux_input_mode
            )
        if output_mode:
            self._set_all_outputs_to(
                module_instance, output_mode, datastore_name=datastore_name
            )
        else:
            self._set_all_outputs_to(
                module_instance,
                self.config.compute.linux_output_mode,
                datastore_name=datastore_name,
            )

    def _apply_scope_runsettings(
        self,
        module_name,
        module_instance: Scope,
        input_mode=None,
        output_mode=None,
        scope_param=None,
        custom_job_name_suffix=None,
        adla_account_name=None,
        priority=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a scope module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            scope_param (str): Parameters to pass to scope e.g. Nebula parameters, VC allocation parameters etc.
            custom_job_name_suffix (str): Optional parameter defining custom string to append to job name
            adla_account_name (str): The name of the Cosmos-migrated Azure Data Lake Analytics account to submit scope job
            custom_runtime_arguments (dict): any additional custom args
        """
        # check if node type is supported
        self._check_node_type(node=module_instance, supported_types=Scope)

        log.info(
            f"Using scope compute target to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unhandled custom_runtime_arguments: {custom_runtime_arguments}"
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(module_instance, output_mode, compliant=False)

        module_instance.adla_account_name = adla_account_name
        module_instance.scope_param = scope_param
        module_instance.custom_job_name_suffix = custom_job_name_suffix
        module_instance.priority = priority

    def _apply_datatransfer_runsettings(
        self,
        module_name,
        module_instance: DataTransfer,
        compliant=True,
        target=None,
        input_mode=None,
        output_mode=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a HDI module.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            compliant (bool): destination datastore, `compliant_datastore` if True, else `noncompliant_datastore`
            target (str): force target compute over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf
            custom_runtime_arguments (dict): any additional custom args
        """
        # check if node type is supported
        self._check_node_type(node=module_instance, supported_types=DataTransfer)

        log.info(
            f"Using datatransfer compute target {self.config.compute.datatransfer_target} to run {module_name} from pipeline class {self.__class__.__name__}"
        )
        if custom_runtime_arguments:
            raise RuntimeError(
                f"Unhandled custom_runtime_arguments: {custom_runtime_arguments}"
            )

        module_instance.compute = (
            target if target is not None else self.config.compute.datatransfer_target
        )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, compliant, datastore_name
        )

    def _apply_sweep_runsettings(
        self,
        module_name,
        module_instance: Sweep,
        windows=False,
        gpu=False,
        target=None,
        input_mode=None,
        output_mode=None,
        node_count=None,
        process_count_per_node=None,
        process_count_per_instance=None,
        algorithm=None,
        # primary_metric=None,
        # goal=None,
        policy_type=None,
        evaluation_interval=None,
        delay_evaluation=None,
        slack_factor=None,
        slack_amount=None,
        truncation_percentage=None,
        max_total_trials=None,
        max_concurrent_trials=None,
        timeout_seconds=None,
        timeout_minutes=None,
        datastore_name=None,
        **custom_runtime_arguments,
    ):
        """Applies settings for a sweep component.

        Args:
            module_name (str): name of the module from the module manifest (required_modules() method)
            module_instance (Module): the AML module we need to add settings to
            target (str): force target compute over hydra conf
            input_mode (str): force input_mode over hydra conf
            output_mode (str): force output_mode over hydra conf

            # For below sweep specific parameters configurations, see below doc link for more info:
            # https://componentsdk.azurewebsites.net/components/sweep_component.html#set-runsettings
            algorithm (str): sweep sampling method
            primary_metric (str): the primary metric of the hyperparameter tuning to optimize
            goal (str): whether the primary metric will be maximize or minimize when evaluating the trials
            policy_type (str): sweep early termination policy type
            evaluation_interval (int): frequency of applying the policy
            delay_evaluation (int): delays the first policy evaluation for a specified number of intervals
            slack_factor (float): the slack allowed with respect to the best performing training run, as a ratio
            slack_amount (float): the slack allowed with respect to the best performing training run, as an absolute ampunt. You should only specify either slack_factor or slack_amount, but not both.
            truncation_percentage (int): the percentage of lowest performing runs to terminate at each evaluation interval. An integer value between 1 and 99.
            max_total_trials (int): maximum number of trial runs. Must be an integer between 1 and 1000.
            max_concurrent_trials (int): maximum number of runs that can run concurrently. If not specified, all runs launch in parallel. If specified, must be an integer between 1 and 100.
            timeout_minutes (int): maximum duration, in minutes, of the hyperparameter tuning experiment. Runs after this duration are canceled.

            custom_runtime_arguments (dict): any additional custom args
        """

        # check if node type is supported
        self._check_node_type(node=module_instance, supported_types=Sweep)

        if target is not None:
            target = target
        elif self.module_loader.is_local(module_name):
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    target = self.config.compute.windows_cpu_dc_target
            else:
                if gpu:
                    target = self.config.compute.linux_gpu_dc_target
                else:
                    target = self.config.compute.linux_cpu_dc_target
        else:
            if windows:
                if gpu:
                    raise ValueError(
                        "A GPU compute target with Windows OS is not available yet!"
                    )
                else:
                    target = self.config.compute.windows_cpu_prod_target
            else:
                if gpu:
                    target = self.config.compute.linux_gpu_prod_target
                else:
                    target = self.config.compute.linux_cpu_prod_target
        log.info(
            f"Using target {target} to run sweep component {module_name} from pipeline class {self.__class__.__name__}"
        )

        module_instance.compute = target

        process_count_per_instance = self._get_process_count_per_node(
            process_count_per_node, process_count_per_instance
        )

        if node_count:
            log.info(
                f"Setting node_count={node_count} and process_count_per_instance={process_count_per_instance} as run settings for sweep component {module_name} from pipeline class {self.__class__.__name__}"
            )

            if module_instance.trial.resources is None:
                module_instance.trial.resources = JobResourceConfiguration()
            module_instance.trial.resources.instance_count = node_count
            if module_instance.trial.distribution is None:
                module_instance.trial.distribution = DistributionConfiguration()
            module_instance.trial.distribution.process_count_per_instance = (
                process_count_per_instance
            )

        if input_mode:
            self._set_all_inputs_to(module_instance, input_mode)
        self._set_all_outputs_to(
            module_instance, output_mode, datastore_name=datastore_name
        )

        if "bandit" in policy_type.lower():
            early_termination_policy = BanditPolicy(
                evaluation_interval=evaluation_interval,
                delay_evaluation=delay_evaluation,
                slack_factor=slack_factor,
                slack_amount=slack_amount,
            )
        if "median" in policy_type.lower():
            early_termination_policy = MedianStoppingPolicy(
                evaluation_interval=evaluation_interval,
                delay_evaluation=delay_evaluation,
            )
        if "trancation" in policy_type.lower():
            early_termination_policy = TruncationSelectionPolicy(
                evaluation_interval=evaluation_interval,
                delay_evaluation=delay_evaluation,
                truncation_percentage=truncation_percentage,
            )
        if timeout_minutes and not timeout_seconds:
            log.warn(
                "Please use timeout_seconds for sweep components, instead of timeout_seconds"
            )
            timeout_seconds = timeout_minutes * 60

        module_instance.sampling_algorithm = algorithm
        # module_instance.objective.primary_metric = primary_metric
        # module_instance.objective.goal = goal
        module_instance.early_termination = early_termination_policy
        module_instance.set_limits(
            timeout=timeout_seconds,
            max_total_trials=max_total_trials,
            max_concurrent_trials=max_concurrent_trials,
        )
        for k, v in custom_runtime_arguments.items():
            setattr(module_instance, k, v)

    def _get_component_name_helper(self, component_instance):
        if isinstance(component_instance, Sweep):
            return component_instance.trial.name
        else:
            return component_instance.component.name

    def _update_dc_configs(
        self, pipeline: PipelineJob, debug: bool = False
    ) -> PipelineJob:
        # collect the known DC targets
        # workaround for missing "throw_on_missing" in currently used omegaconf
        compute_config_as_dict = yaml.safe_load(OmegaConf.to_yaml(self.config.compute))
        dc_targets: Set[str] = {
            target
            for key, target in compute_config_as_dict.items()
            if key.endswith("_dc_target") and target and target != str(MISSING)
        }
        dc_datastore = self.config.compute.dc_datastore
        if not dc_targets:
            # This is kind of pointless because if one local component exists
            # then one of the `apply_*` methods wil fail if no dc target is not
            # specified. And if no local component exists, then there is no
            # point in outputting this. But let's do it anyway:
            log.info("The compute config does not contain any DC targets.")
            return pipeline

        log.info("Overriding compute target if upstream steps use DC...")

        # build and parse the dependency graph
        dependency_graph = _DependencyGraph(dc_targets)

        should_use_dc = dependency_graph.determine_dc_nodes(pipeline)
        log.info(f"Need to configure the following for DC: {sorted(should_use_dc)}")

        target_map: Dict[str, str] = {}
        for prod_target in [
            "hdi_prod_target",
            "linux_cpu_prod_target",
            "linux_gpu_prod_target",
            "windows_cpu_prod_target",
            "synapse_prod_target",
        ]:
            if prod_target not in self.config.compute:
                continue
            target_map[self.config.compute[prod_target]] = self.config.compute.get(
                prod_target.replace("_prod_", "_dc_"), None
            )

        for id in should_use_dc:
            node = dependency_graph.id_2_node[id]
            if str(node.type) == "ScopeComponent":
                # scope component doesn't need to configure compute target
                # output datastore has to be cosmos (ADLSg1), while dc_datastore is ADLSg2
                continue
            compute_property_name = (
                "compute_name" if str(node.type) == "HDInsightComponent" else "compute"
            )
            current_compute_target = getattr(node, compute_property_name)
            if (
                isinstance(current_compute_target, str)
                and current_compute_target in target_map
            ):
                setattr(node, compute_property_name, target_map[current_compute_target])
                log.info(
                    f"Updating compute target for {node.name} to {target_map[current_compute_target]}"
                )
            else:
                log.error(
                    f"Node {node} is using an unknown target: "
                    f"{current_compute_target}. "
                    f"Please move this to DC manually, or re-run "
                    f"this after adding the target to the compute config."
                )
            self._set_all_outputs_to(node, datastore_name=dc_datastore)
            log.info(f"Writing outputs of {node.name} to dc_datastore {dc_datastore}")
        return pipeline


class _DependencyGraph:
    # modified from https://github.com/Azure/DesignerPrivatePreviewFeatures/blob/master/azure-ai-ml/samples/inputs_and_outputs/det_ch_sample/pipeline.ipynb

    def __init__(self, dc_targets):
        self.dc_targets = dc_targets
        self.id_2_node = {}
        self.id_2_dc = {}

    def determine_dc_nodes(self, pipeline):
        """Determine which nodes should use dc."""
        should_use_dc = set()
        graph = self.build_graph(pipeline)
        for node_id in toposort_flatten(graph):
            node = self.id_2_node[node_id]
            if self.should_use_dc(node, graph[node_id]):
                should_use_dc.add(node_id)
        return should_use_dc

    def get_predecessors(self, node) -> List[BaseNode]:
        """Return list of predecessors for current node.

        Note: Only non-control flow nodes in @pipeline are supported.
        Node: For sub-graph node, we will trace back to inner node and return.
        Example:
            @pipeline
            def sub_pipeline():
                inner_node = component_func()
                return inner_node.outputs
            @pipeline
            def root_pipeline():
                pipeline_node = sub_pipeline()
                node1 = component_func(input1=pipeline_node.outputs.output1)
                node2 = component_func(
                    input1=pipeline_node.outputs.output1
                    input2=node1.outputs.output1
                )
                # pipeline_node.self.get_predecessors() will return []
                # node1.self.get_predecessors() will return [inner_node]
                # node2.self.get_predecessors() will return [inner_node, node1]
        """

        # use {id: instance} dict to avoid nodes with component and parameters being duplicated
        predecessors = {}
        for _, input_value in node.inputs.items():
            if not isinstance(input_value, NodeInput):
                continue
            owner = input_value._get_data_owner()
            if owner is not None:
                predecessors[owner._instance_id] = owner
        return list(predecessors.values())

    def expand_pipeline_nodes(self, pipeline: Union[PipelineJob, PipelineComponent]):
        """Expand pipeline nodes to a list of nodes. All sub-graphs will be expanded."""
        nodes = []
        for node in pipeline.jobs.values():
            if isinstance(node, Pipeline):
                pipeline_component = node.component
                if not isinstance(pipeline_component, PipelineComponent):
                    raise RuntimeError(
                        "Pipeline component must be a PipelineComponent object, but got {}".format(
                            type(pipeline_component)
                        )
                    )
                nodes.extend(self.expand_pipeline_nodes(pipeline_component))
            else:
                nodes.append(node)
        return nodes

    def map_to_id(self, item: BaseNode) -> Optional[str]:
        """Maps the given item to an id."""
        if isinstance(item, BaseNode):
            # directly use node instance id
            node_id = item._instance_id
            self.id_2_node[node_id] = item
            return node_id
        raise ValueError(
            f"Please report this. Could not map {item} to dependency graph id."
        )

    def build_graph(self, pipeline):
        """Build a graph to toposort"""
        graph = {}

        nodes = self.expand_pipeline_nodes(pipeline)

        for node in nodes:
            node_id = self.map_to_id(node)
            predecessors = self.get_predecessors(node)
            for predecessor in predecessors:
                predecessor_id = self.map_to_id(predecessor)
                if node_id not in graph:
                    graph[node_id] = set()
                if predecessor_id not in graph:
                    graph[predecessor_id] = set()
                graph[node_id].add(predecessor_id)

        return graph

    def should_use_dc(self, node, predecessors_ids):
        """Return True if a node should use dc."""
        node_id = self.map_to_id(node)
        # If specified dc compute, use dc
        if (
            hasattr(node, "compute")
            and isinstance(node.compute, str)
            and node.compute in self.dc_targets
        ):
            self.id_2_dc[node_id] = True
            return True
        elif (
            hasattr(node, "compute_name")
            and isinstance(node.compute_name, str)
            and node.compute_name in self.dc_targets
        ):
            self.id_2_dc[node_id] = True
            return True
        # Otherwise, determine by its predecessors
        for predecessor_id in predecessors_ids:
            if self.id_2_dc.get(predecessor_id, False):
                self.id_2_dc[node_id] = True
                return True
        self.id_2_dc[node_id] = False
        return False



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
"""A dummy module for backward compatability for code like e.g.
`from shrike.pipeline.argparser_utils import get_args_from_component_spec`.
Until `shrike<=1.34.0`, `argparser_utils` module existed under `shrike.pipeline`.
"""
from shrike.compliant_logging.argparser_utils import *



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.


class ShrikeException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)


class ShrikeUserErrorException(ShrikeException):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Python utilities to aid "compliant experimentation" - training
machine learning models without seeing the training data.
"""

__version__ = "2.0.0.dev16"

import os

# enable internal components in v2
os.environ["AZURE_ML_INTERNAL_COMPONENTS_ENABLED"] = "True"
os.environ["AZURE_ML_CLI_PRIVATE_FEATURES_ENABLED"] = "True"



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.
from azure.ai.ml.entities import Environment, CommandJob
import os
from omegaconf import OmegaConf
import hydra
from hydra.core.hydra_config import HydraConfig
from shrike.pipeline.aml_connect import azureml_connect
from shrike._core import (
    is_eyesoff_helper,
    b64_decode,
    b64_encode,
    experimental,
    O365_FEED,
)
from shrike.pipeline import AMLPipelineHelper


def merge_serialize_config_with_default(serialized_config):
    default_config = AMLPipelineHelper._default_config()
    cfg = OmegaConf.merge(
        default_config, OmegaConf.structured(b64_decode(serialized_config))
    )
    return cfg


def get_or_create_env(ml_client, env_name, is_eyesoff):
    try:
        env = ml_client.environemnts.get(name=env_name)
    except Exception:
        pip_packages = [
            "ray>=1.7.0",
            "shrike[pipeline]>=1.28.0",
        ]
        if is_eyesoff:
            pip_packages.append(f"--index-url {O365_FEED}")
        conda_file_dict = {
            "name": env_name,
            "dependencies": [
                "python=3.8",
                "pip",
                {
                    "pip": pip_packages,
                },
            ],
        }
        env = Environment(
            name=env_name,
            conda_file=conda_file_dict,
        )

    return env


@experimental()
@hydra.main(config_name="default")
def ray_actor_on_shrike(cfg):
    default_config = AMLPipelineHelper._default_config()
    cfg = OmegaConf.merge(default_config, cfg)
    env_name = "ray_env"

    aml = cfg.aml
    ml_client = azureml_connect(
        aml_subscription_id=aml.subscription_id,
        aml_resource_group=aml.resource_group,
        aml_workspace_name=aml.workspace_name,
        aml_auth=aml.get("auth", "interactive"),
        aml_tenant=aml.get("tenant", None),
        aml_force=aml.get("force", False),
    )
    env = get_or_create_env(
        ml_client,
        env_name,
        is_eyesoff_helper(aml.get("tenant", None), aml.subscription_id),
    )
    compute_target = cfg.compute.default_compute_target

    script = os.path.join(HydraConfig.get().runtime.cwd, cfg.run.script)
    with open(script) as file:
        contents = file.read()

    command_job = CommandJob(
        code=".",
        command=f"echo '{contents}' > script.py && python script.py --serialized_config {b64_encode(OmegaConf.to_yaml(cfg))}",
        compute=compute_target,
        environment=env,
    )
    job = ml_client.jobs.create_or_update(
        command_job, experiment_name=cfg.run.experiment_name
    )
    print(job.studio_url)
    ml_client.jobs.stream(job.name)
